"""
ART-specific data structures

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: Christopher Moody <cemoody@ucsc.edu>
Affiliation: UCSC
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import os.path
import glob
import stat
import weakref
import cStringIO

from yt.funcs import *
from yt.data_objects.grid_patch import \
      AMRGridPatch
from yt.data_objects.hierarchy import \
      AMRHierarchy
from yt.data_objects.static_output import \
      StaticOutput
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from .fields import \
    ARTFieldInfo, add_art_field, KnownARTFields
from yt.utilities.definitions import \
    mpc_conversion
from yt.utilities.io_handler import \
    io_registry
from yt.utilities.lib import \
    get_box_grids_level
import yt.utilities.lib as amr_utils

from .definitions import *
from io import _read_child_mask_level
from io import read_particles
from io import read_stars
from io import spread_ages
from io import _count_art_octs
from io import _read_art_level_info
from io import _read_art_child
from io import _skip_record
from io import _read_record
from io import _read_frecord
from io import _read_record_size
from io import _read_struct
from io import b2t


import yt.frontends.ramses._ramses_reader as _ramses_reader

from .fields import ARTFieldInfo, KnownARTFields
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion
from yt.utilities.lib import \
    get_box_grids_level
from yt.utilities.io_handler import \
    io_registry
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from yt.utilities.physical_constants import \
    mass_hydrogen_cgs, sec_per_Gyr

class ARTGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, hierarchy, level, locations,start_index, le,re,gd,
            child_mask=None,nop=0):
        AMRGridPatch.__init__(self, id, filename = hierarchy.hierarchy_filename,
                              hierarchy = hierarchy)
        start_index =start_index 
        self.Level = level
        self.Parent = []
        self.Children = []
        self.locations = locations
        self.start_index = start_index.copy()
        
        self.LeftEdge = le
        self.RightEdge = re
        self.ActiveDimensions = gd
        self.NumberOfParticles=nop
        for particle_field in particle_fields:
            setattr(self,particle_field,np.array([]))

    def _setup_dx(self):
        id = self.id - self._id_offset
        if len(self.Parent) > 0:
            self.dds = self.Parent[0].dds / self.pf.refine_by
        else:
            LE, RE = self.hierarchy.grid_left_edge[id,:], \
                     self.hierarchy.grid_right_edge[id,:]
            self.dds = np.array((RE-LE)/self.ActiveDimensions)
        if self.pf.dimensionality < 2: self.dds[1] = 1.0
        if self.pf.dimensionality < 3: self.dds[2] = 1.0
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] \
                = self.dds

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.
        """
        if self.start_index != None:
            return self.start_index
        if len(self.Parent) == 0:
            start_index = self.LeftEdge / self.dds
            return np.rint(start_index).astype('int64').ravel()
        pdx = self.Parent[0].dds
        start_index = (self.Parent[0].get_global_startindex()) + \
                       np.rint((self.LeftEdge - self.Parent[0].LeftEdge)/pdx)
        self.start_index = (start_index*self.pf.refine_by)\
                           .astype('int64').ravel()
        return self.start_index

    def __repr__(self):
        return "ARTGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

class ARTHierarchy(AMRHierarchy):
    grid = ARTGrid
    _handle = None
    
    def __init__(self, pf, data_style='art'):
        self.data_style = data_style
        self.parameter_file = weakref.proxy(pf)
        self.max_level = pf.max_level
        self.hierarchy_filename = self.parameter_file.parameter_filename
        self.directory = os.path.dirname(self.hierarchy_filename)
        self.float_type = np.float64
        AMRHierarchy.__init__(self,pf,data_style)
        self._setup_field_list()

    def _initialize_data_storage(self):
        pass
    
    def _detect_fields(self):
        self.field_list = []
        self.field_list += fluid_fields
        self.field_list += particle_fields
        
    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()
            
    def _count_grids(self):
        LEVEL_OF_EDGE = 7
        MAX_EDGE = (2 << (LEVEL_OF_EDGE- 1))
        min_eff = 0.30
        vol_max = 128**3
        with open(self.pf.parameter_filename,'rb') as f:
            (self.pf.nhydro_vars, self.pf.level_info,
            self.pf.level_oct_offsets, 
            self.pf.level_child_offsets) = \
                             _count_art_octs(f, 
                              self.pf.child_grid_offset,
                              self.pf.min_level, self.pf.max_level)
            self.pf.level_info[0]=self.pf.ncell
            self.pf.level_info = np.array(self.pf.level_info)
            self.pf.level_offsets = self.pf.level_child_offsets
            self.pf.level_offsets = np.array(self.pf.level_offsets, 
                                             dtype='int64')
            self.pf.level_offsets[0] = self.pf.root_grid_offset
            self.pf.level_art_child_masks = {}
            cm = self.pf.root_iOctCh>0
            cm_shape = (1,)+cm.shape 
            self.pf.level_art_child_masks[0] = \
                    cm.reshape(cm_shape).astype('uint8')        
            del cm
            root_psg = _ramses_reader.ProtoSubgrid(
                            np.zeros(3, dtype='int64'), # left index of PSG
                            self.pf.domain_dimensions, # dim of PSG
                            np.zeros((1,3), dtype='int64'),# left edges of grids
                            np.zeros((1,6), dtype='int64') # empty
                            )
            self.proto_grids = [[root_psg],]
            for level in xrange(1, len(self.pf.level_info)):
                if self.pf.level_info[level] == 0:
                    self.proto_grids.append([])
                    continue
                psgs = []
                effs,sizes = [], []
                if self.pf.limit_level:
                    if level > self.pf.limit_level : continue
                #refers to the left index for the art octgrid
                left_index, fl, nocts,root_level = _read_art_level_info(f, 
                        self.pf.level_oct_offsets,level,
                        coarse_grid=self.pf.domain_dimensions[0])
                if level>1:
                    assert root_level == last_root_level
                last_root_level = root_level
                #left_index_gridpatch = left_index >> LEVEL_OF_EDGE
                #read in the child masks for this level and save them
                idc, art_child_mask = _read_child_mask_level(f, 
                        self.pf.level_child_offsets,
                    level,nocts*8,nhydro_vars=self.pf.nhydro_vars)
                art_child_mask = art_child_mask.reshape((nocts,2,2,2))
                self.pf.level_art_child_masks[level]=art_child_mask
                #child_mask is zero where child grids exist and
                #thus where higher resolution data is available
                #compute the hilbert indices up to a certain level
                #the indices will associate an oct grid to the nearest
                #hilbert index?
                base_level = int( np.log10(self.pf.domain_dimensions.max()) /
                                  np.log10(2))
                hilbert_indices = _ramses_reader.get_hilbert_indices(
                                        level + base_level, left_index)
                #print base_level, hilbert_indices.max(),
                hilbert_indices = hilbert_indices >> base_level + LEVEL_OF_EDGE
                #print hilbert_indices.max()
                # Strictly speaking, we don't care about the index of any
                # individual oct at this point.  So we can then split them up.
                unique_indices = np.unique(hilbert_indices)
                mylog.info("Level % 2i has % 10i unique indices for %0.3e octs",
                            level, unique_indices.size, hilbert_indices.size)
                #use the hilbert indices to order oct grids so that consecutive
                #items on a list are spatially near each other
                #this is useful because we will define grid patches over these
                #octs, which are more efficient if the octs are spatially close
                #split into list of lists, with domains containing 
                #lists of sub octgrid left indices and an index
                #referring to the domain on which they live
                pbar = get_pbar("Calc Hilbert Indices ",1)
                locs, lefts = _ramses_reader.get_array_indices_lists(
                            hilbert_indices, unique_indices, left_index, fl)
                pbar.finish()
                #iterate over the domains    
                step=0
                pbar = get_pbar("Re-gridding  Level %i "%level, len(locs))
                psg_eff = []
                for ddleft_index, ddfl in zip(lefts, locs):
                    #iterate over just the unique octs
                    #why would we ever have non-unique octs?
                    #perhaps the hilbert ordering may visit the same
                    #oct multiple times - review only unique octs 
                    #for idomain in np.unique(ddfl[:,1]):
                    #dom_ind = ddfl[:,1] == idomain
                    #dleft_index = ddleft_index[dom_ind,:]
                    #dfl = ddfl[dom_ind,:]
                    dleft_index = ddleft_index
                    dfl = ddfl
                    initial_left = np.min(dleft_index, axis=0)
                    idims = (np.max(dleft_index, axis=0) - initial_left).ravel()
                    idims +=2
                    #this creates a grid patch that doesn't cover the whole leve
                    #necessarily, but with other patches covers all the regions
                    #with octs. This object automatically shrinks its size
                    #to barely encompass the octs inside of it.
                    psg = _ramses_reader.ProtoSubgrid(initial_left, idims,
                                    dleft_index, dfl)
                    if psg.efficiency <= 0: continue
                    #because grid patches maybe mostly empty, and with octs
                    #that only partially fill the grid, it may be more efficient
                    #to split large patches into smaller patches. We split
                    #if less than 10% the volume of a patch is covered with octs
                    if idims.prod() > vol_max or psg.efficiency < min_eff:
                        psg_split = _ramses_reader.recursive_patch_splitting(
                            psg, idims, initial_left, 
                            dleft_index, dfl,min_eff=min_eff,use_center=True,
                            split_on_vol=vol_max)
                        psgs.extend(psg_split)
                        psg_eff += [x.efficiency for x in psg_split] 
                    else:
                        psgs.append(psg)
                        psg_eff =  [psg.efficiency,]
                    tol = 1.00001
                    step+=1
                    pbar.update(step)
                eff_mean = np.mean(psg_eff)
                eff_nmin = np.sum([e<=min_eff*tol for e in psg_eff])
                eff_nall = len(psg_eff)
                mylog.info("Average subgrid efficiency %02.1f %%",
                            eff_mean*100.0)
                mylog.info("%02.1f%% (%i/%i) of grids had minimum efficiency",
                            eff_nmin*100.0/eff_nall,eff_nmin,eff_nall)
                
                mylog.info("Done with level % 2i; max LE %i", level,
                           np.max(left_index))
                pbar.finish()
                self.proto_grids.append(psgs)
                if len(self.proto_grids[level]) == 1: continue
        self.num_grids = sum(len(l) for l in self.proto_grids)
        
    def _parse_hierarchy(self):
        grids = []
        gi = 0
        dd=self.pf.domain_dimensions
        for level, grid_list in enumerate(self.proto_grids):
            dds = ((2**level) * dd).astype("float64")
            for g in grid_list:
                fl = g.grid_file_locations
                props = g.get_properties()
                start_index = props[0,:]
                le = props[0,:].astype('float64')/dds
                re = props[1,:].astype('float64')/dds
                gd = props[2,:].astype('int64')
                if level==0:
                    le = np.zeros(3,dtype='float64')
                    re = np.ones(3,dtype='float64')
                    gd = dd
                self.grid_left_edge[gi,:] = le
                self.grid_right_edge[gi,:] = re
                self.grid_dimensions[gi,:] = gd
                assert np.all(self.grid_left_edge[gi,:]<=1.0)    
                self.grid_levels[gi,:] = level
                child_mask = np.zeros(props[2,:],'uint8')
                amr_utils.fill_child_mask(fl,start_index,
                    self.pf.level_art_child_masks[level],
                    child_mask)
                grids.append(self.grid(gi, self, level, fl, 
                    start_index,le,re,gd))
                gi += 1
        self.grids = np.empty(len(grids), dtype='object')
        if not self.pf.skip_particles and self.pf.file_particle_data:
            lspecies = self.pf.parameters['lspecies']
            wspecies = self.pf.parameters['wspecies']
            self.pf.particle_position,self.pf.particle_velocity= \
                read_particles(self.pf.file_particle_data,
                        self.pf.parameters['Nrow'])
            nparticles = lspecies[-1]
            if not np.all(self.pf.particle_position[nparticles:]==0.0):
                mylog.info('WARNING: unused particles discovered from lspecies')
            self.pf.particle_position = self.pf.particle_position[:nparticles]
            self.pf.particle_velocity = self.pf.particle_velocity[:nparticles]
            self.pf.particle_position  /= self.pf.domain_dimensions 
            self.pf.particle_type = np.zeros(nparticles,dtype='int')
            self.pf.particle_mass = np.zeros(nparticles,dtype='float64')
            self.pf.particle_star_index = len(wspecies)-1
            a=0
            for i,(b,m) in enumerate(zip(lspecies,wspecies)):
                if i == self.pf.particle_star_index:
                    assert m==0.0
                    sa,sb = a,b
                else:
                    assert m>0.0
                self.pf.particle_type[a:b] = i #particle type
                self.pf.particle_mass[a:b] = m #mass in code units
                a=b
            if not self.pf.skip_stars and self.pf.file_particle_stars: 
                (nstars_rs,), mass, imass, tbirth, metallicity1, metallicity2, \
                        ws_old,ws_oldi,tdum,adum \
                     = read_stars(self.pf.file_particle_stars)
                self.pf.nstars_rs = nstars_rs     
                self.pf.nstars_pa = b-a
                inconsistent=self.pf.particle_type==self.pf.particle_star_index
                if not nstars_rs==np.sum(inconsistent):
                    mylog.info('WARNING!: nstars is inconsistent!')
                del inconsistent
                if nstars_rs > 0 :
                    n=min(1e2,len(tbirth))
                    birthtimes= b2t(tbirth,n=n)
                    birthtimes = birthtimes.astype('float64')
                    assert birthtimes.shape == tbirth.shape    
                    birthtimes*= 1.0e9 #from Gyr to yr
                    birthtimes*= 365*24*3600 #to seconds
                    ages = self.pf.current_time-birthtimes
                    spread = self.pf.spread_age
                    if type(spread)==type(5.5):
                        ages = spread_ages(ages,spread=spread)
                    elif spread:
                        ages = spread_ages(ages)
                    idx = self.pf.particle_type == self.pf.particle_star_index
                    for psf in particle_star_fields:
                        if getattr(self.pf,psf,None) is None:
                            setattr(self.pf,psf,
                                    np.zeros(nparticles,dtype='float64'))
                    self.pf.particle_age[sa:sb] = ages
                    self.pf.particle_mass[sa:sb] = mass
                    self.pf.particle_mass_initial[sa:sb] = imass
                    self.pf.particle_creation_time[sa:sb] = birthtimes
                    self.pf.particle_metallicity1[sa:sb] = metallicity1
                    self.pf.particle_metallicity2[sa:sb] = metallicity2
                    self.pf.particle_metallicity[sa:sb]  = metallicity1\
                                                          + metallicity2
        for gi,g in enumerate(grids):    
            self.grids[gi]=g
                    
    def _get_grid_parents(self, grid, LE, RE):
        mask = np.zeros(self.num_grids, dtype='bool')
        grids, grid_ind = self.get_box_grids(LE, RE)
        mask[grid_ind] = True
        mask = np.logical_and(mask, (self.grid_levels == (grid.Level-1)).flat)
        return self.grids[mask]

    def _populate_grid_objects(self):
        mask = np.empty(self.grids.size, dtype='int32')
        pb = get_pbar("Populating grids", len(self.grids))
        for gi,g in enumerate(self.grids):
            pb.update(gi)
            amr_utils.get_box_grids_level(self.grid_left_edge[gi,:],
                                self.grid_right_edge[gi,:],
                                g.Level - 1,
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask)
            parents = self.grids[mask.astype("bool")]
            if len(parents) > 0:
                g.Parent.extend((p for p in parents.tolist()
                        if p.locations[0,0] == g.locations[0,0]))
                for p in parents: p.Children.append(g)
            #Now we do overlapping siblings; note that one has to "win" with
            #siblings, so we assume the lower ID one will "win"
            amr_utils.get_box_grids_level(self.grid_left_edge[gi,:],
                                self.grid_right_edge[gi,:],
                                g.Level,
                                self.grid_left_edge, self.grid_right_edge,
                                self.grid_levels, mask, gi)
            mask[gi] = False
            siblings = self.grids[mask.astype("bool")]
            if len(siblings) > 0:
                g.OverlappingSiblings = siblings.tolist()
            g._prepare_grid()
            g._setup_dx()
            #instead of gridding particles assign them all to the root grid
            if gi==0:
                for particle_field in particle_fields:
                    source = getattr(self.pf,particle_field,None)
                    if source is None:
                        for i,ax in enumerate('xyz'):
                            pf = particle_field.replace('_%s'%ax,'')
                            source = getattr(self.pf,pf,None)
                            if source is not None:
                                source = source[:,i]
                                break
                    if source is not None:
                        mylog.info("Attaching %s to the root grid",
                                    particle_field)
                        g.NumberOfParticles = source.shape[0]
                        setattr(g,particle_field,source)
                g.particle_index = np.arange(g.NumberOfParticles)
        pb.finish()
        self.max_level = self.grid_levels.max()

    def _setup_field_list(self):
        if not self.parameter_file.skip_particles:
            # We know which particle fields will exist -- pending further
            # changes in the future.
            for field in particle_fields:
                def external_wrapper(f):
                    def _convert_function(data):
                        return data.convert(f)
                    return _convert_function
                cf = external_wrapper(field)
                # Note that we call add_field on the field_info directly.  This
                # will allow the same field detection mechanism to work for 1D,
                # 2D and 3D fields.
                self.pf.field_info.add_field(field, NullFunc,
                                             convert_function=cf,
                                             take_log=True, particle_type=True)

    def _setup_derived_fields(self):
        self.derived_field_list = []

    def _setup_data_io(self):
        self.io = io_registry[self.data_style](
            self.pf.parameter_filename,
            self.pf.nhydro_vars,
            self.pf.level_info,
            self.pf.level_offsets)

class ARTStaticOutput(StaticOutput):
    _hierarchy_class = ARTHierarchy
    _fieldinfo_fallback = ARTFieldInfo
    _fieldinfo_known = KnownARTFields
    
    def __init__(self, file_amr, storage_filename = None,
            skip_particles=False,skip_stars=False,limit_level=None,
            spread_age=True,data_style='art'):
        self.data_style = data_style
        self._find_files(file_amr)
        self.skip_particles = skip_particles
        self.skip_stars = skip_stars
        self.file_amr = file_amr
        self.parameter_filename = file_amr
        self.limit_level = limit_level
        self.spread_age = spread_age
        self.domain_left_edge  = np.zeros(3,dtype='float64')
        self.domain_right_edge = np.ones(3,dtype='float64') 
        StaticOutput.__init__(self, file_amr, data_style)
        self.storage_filename = storage_filename

    def _find_files(self,file_amr):
        """
        Given the AMR base filename, attempt to find the
        particle header, star files, etc.
        """
        prefix,suffix = filename_pattern['amr'].split('%s')
        affix = os.path.basename(file_amr).replace(prefix,'')
        affix = affix.replace(suffix,'')
        affix = affix.replace('_','')
        affix = affix[1:-1]
        dirname = os.path.dirname(file_amr)
        for filetype, pattern in filename_pattern.items():
            #sometimes the affix is surrounded by an extraneous _
            #so check for an extra character on either side
            check_filename = dirname+'/'+pattern%('?%s?'%affix)
            filenames = glob.glob(check_filename)
            if len(filenames)==1:
                setattr(self,"file_"+filetype,filenames[0])
                mylog.info('discovered %s',filetype)
            elif len(filenames)>1:
                setattr(self,"file_"+filetype,None)
                mylog.info("Ambiguous number of files found for %s",
                        check_filename)
            else:
                setattr(self,"file_"+filetype,None)

    def __repr__(self):
        return self.basename.rsplit(".", 1)[0]
        
    def _set_units(self):
        """
        Generates the conversion to various physical units based 
		on the parameters from the header
        """
        self.units = {}
        self.time_units = {}
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['unitary'] = 1.0

        #spatial units
        z   = self.current_redshift
        h   = self.hubble_constant
        boxcm_cal = self.parameters["boxh"]
        boxcm_uncal = boxcm_cal / h
        box_proper = boxcm_uncal/(1+z)
        aexpn = self["aexpn"]
        for unit in mpc_conversion:
            self.units[unit] = mpc_conversion[unit] * box_proper
            self.units[unit+'h'] = mpc_conversion[unit] * box_proper * h
            self.units[unit+'cm'] = mpc_conversion[unit] * boxcm_uncal
            self.units[unit+'hcm'] = mpc_conversion[unit] * boxcm_cal

        #all other units
        wmu = self.parameters["wmu"]
        Om0 = self.parameters['Om0']
        ng  = self.parameters['ng']
        wmu = self.parameters["wmu"]
        boxh   = self.parameters['boxh'] 
        aexpn  = self.parameters["aexpn"]
        hubble = self.parameters['hubble']

        cf = defaultdict(lambda: 1.0)
        r0 = boxh/ng
        P0= 4.697e-16 * Om0**2.0 * r0**2.0 * hubble**2.0
        T_0 = 3.03e5 * r0**2.0 * wmu * Om0 # [K]
        S_0 = 52.077 * wmu**(5.0/3.0)
        S_0 *= hubble**(-4.0/3.0)*Om0**(1.0/3.0)*r0**2.0
        #v0 =  r0 * 50.0*1.0e5 * np.sqrt(self.omega_matter)  #cm/s
        v0 = 50.0*r0*np.sqrt(Om0)
        t0 = r0/v0
        #rho0 = 1.8791e-29 * hubble**2.0 * self.omega_matter
        rho0 = 2.776e11 * hubble**2.0 * Om0
        tr = 2./3. *(3.03e5*r0**2.0*wmu*self.omega_matter)*(1.0/(aexpn**2))     
        aM0 = rho0 * (boxh/hubble)**3.0 / ng**3.0
        cf['r0']=r0
        cf['P0']=P0
        cf['T_0']=T_0
        cf['S_0']=S_0
        cf['v0']=v0
        cf['t0']=t0
        cf['rho0']=rho0
        cf['tr']=tr
        cf['aM0']=aM0

        #factors to multiply the native code units to CGS
        cf['Pressure'] = P0 #already cgs
        cf['Velocity'] = v0/aexpn*1.0e5 #proper cm/s
        cf["Mass"] = aM0 * 1.98892e33
        cf["Density"] = rho0*(aexpn**-3.0)
        cf["GasEnergy"] = rho0*v0**2*(aexpn**-5.0)
        cf["Potential"] = 1.0
        cf["Entropy"] = S_0
        cf["Temperature"] = tr
        self.cosmological_simulation = True
        self.conversion_factors = cf
        
        for particle_field in particle_fields:
            self.conversion_factors[particle_field] =  1.0
        for ax in 'xyz':
            self.conversion_factors["%s-velocity" % ax] = 1.0
            self.conversion_factors["particle_velocity_%s"%ax] = cf['Velocity']
        for unit in sec_conversion.keys():
            self.time_units[unit] = 1.0 / sec_conversion[unit]
        self.conversion_factors['particle_mass'] = cf['Mass']
        self.conversion_factors['particle_creation_time'] =  31556926.0
        self.conversion_factors['Msun'] = 5.027e-34 

    def _parse_parameter_file(self):
        """
        Get the various simulation parameters & constants.
        """
        self.dimensionality = 3
        self.refine_by = 2
        self.periodicity = (True, True, True)
        self.cosmological_simulation = True
        self.parameters = {}
        self.unique_identifier = \
            int(os.stat(self.parameter_filename)[stat.ST_CTIME])
        self.parameters.update(constants)
        with open(self.file_amr,'rb') as f:
            amr_header_vals = _read_struct(f,amr_header_struct)
            for to_skip in ['tl','dtl','tlold','dtlold','iSO']:
                _skip_record(f)
            (self.ncell,) = struct.unpack('>l', _read_record(f))
            # Try to figure out the root grid dimensions
            est = int(np.rint(self.ncell**(1.0/3.0)))
            # Note here: this is the number of *cells* on the root grid.
            # This is not the same as the number of Octs.
            self.domain_dimensions = np.ones(3, dtype='int64')*est 
            self.root_grid_mask_offset = f.tell()
            root_cells = self.domain_dimensions.prod()
            self.root_iOctCh = _read_frecord(f,'>i')[:root_cells]
            self.root_iOctCh = self.root_iOctCh.reshape(self.domain_dimensions,
                 order='F')
            self.root_grid_offset = f.tell()
            _skip_record(f) # hvar
            _skip_record(f) # var
            self.iOctFree, self.nOct = struct.unpack('>ii', _read_record(f))
            self.child_grid_offset = f.tell()
        self.parameters.update(amr_header_vals)
        if not self.skip_particles and self.file_particle_header:
            with open(self.file_particle_header,"rb") as fh:
                particle_header_vals = _read_struct(fh,particle_header_struct)
                fh.seek(seek_extras)
                n = particle_header_vals['Nspecies']
                wspecies = np.fromfile(fh,dtype='>f',count=10)
                lspecies = np.fromfile(fh,dtype='>i',count=10)
            self.parameters['wspecies'] = wspecies[:n]
            self.parameters['lspecies'] = lspecies[:n]
            ls_nonzero = np.diff(lspecies)[:n-1]
            mylog.info("Discovered %i species of particles",len(ls_nonzero))
            mylog.info("Particle populations: "+'%1.1e '*len(ls_nonzero),
                *ls_nonzero)
            for k,v in particle_header_vals.items():
                if k in self.parameters.keys():
                    if not self.parameters[k] == v:
                        mylog.info("Inconsistent parameter %s %1.1e  %1.1e",k,v,
                                   self.parameters[k])
                else:
                    self.parameters[k]=v
            self.parameters_particles = particle_header_vals
    
        #setup standard simulation yt expects to see
        self.current_redshift = self.parameters["aexpn"]**-1.0 - 1.0
        self.omega_lambda = amr_header_vals['Oml0']
        self.omega_matter = amr_header_vals['Om0']
        self.hubble_constant = amr_header_vals['hubble']
        self.min_level = amr_header_vals['min_level']
        self.max_level = amr_header_vals['max_level']
        self.hubble_time  = 1.0/(self.hubble_constant*100/3.08568025e19)
        self.current_time = b2t(self.parameters['t']) * sec_per_Gyr

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Defined for the NMSU file naming scheme.
        This could differ for other formats.
        """
        fn = ("%s" % (os.path.basename(args[0])))
        f = ("%s" % args[0])
        if fn.endswith(".d") and fn.startswith('10Mpc') and\
                os.path.exists(f): 
                return True
        return False




"""
A light interface to a few HEALPix routines

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython

from libc.stdio cimport fopen, fclose, FILE

cdef extern from "healpix_vectors.h":
    int pix2vec_nest(long nside, long ipix, double *v)
    void vec2pix_nest(long nside, double *vec, long *ipix)
    void pix2ang_nest(long nside, long ipix, double *theta, double *phi)
    void ang2pix_nest(long nside, double theta, double phi, long *ipix)



"""
Default tests

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *
from output_tests import YTStaticOutputTest, create_test

class TestFieldStatistics(YTStaticOutputTest):

    tolerance = None

    def run(self):
        # We're going to calculate the field statistics for every single field.
        results = {}
        for field in self.pf.h.field_list:
            # Do it here so that it gets wiped each iteration
            dd = self.pf.h.all_data() 
            results[field] = (dd[field].std(),
                              dd[field].mean(),
                              dd[field].min(),
                              dd[field].max())
        self.result = results

    def compare(self, old_result):
        for field in sorted(self.result):
            for i in range(4):
                oi = old_result[field][i]
                ni = self.result[field][i]
                self.compare_value_delta(oi, ni, self.tolerance)

class TestAllProjections(YTStaticOutputTest):

    tolerance = None

    def run(self):
        results = {}
        for field in self.pf.h.field_list:
            if self.pf.field_info[field].particle_type: continue
            results[field] = []
            for ax in range(3):
                t = self.pf.h.proj(ax, field)
                results[field].append(t.field_data)
        self.result = results

    def compare(self, old_result):
        for field in sorted(self.result):
            for p1, p2 in zip(self.result[field], old_result[field]):
                self.compare_data_arrays(p1, p2, self.tolerance)




"""
This code was modified from the nose plugins/xunit.py module, licensed under
the GNU Lesser General Public License and copyrighted by Jason Pellerin.

For more information, see the nose homepage at:

    http://python-nose.googlecode.com/

--
This plugin provides test results in the standard XUnit XML format.

It was designed for the `Hudson`_ continuous build system but will
probably work for anything else that understands an XUnit-formatted XML
representation of test results.

Add this shell command to your builder ::

    nosetests --with-xunit

And by default a file named nosetests.xml will be written to the
working directory.

In a Hudson builder, tick the box named "Publish JUnit test result report"
under the Post-build Actions and enter this value for Test report XMLs::

    **/nosetests.xml

If you need to change the name or location of the file, you can set the
``--xunit-file`` option.

Here is an abbreviated version of what an XML test report might look like::

    <?xml version="1.0" encoding="UTF-8"?>
    <testsuite name="nosetests" tests="1" errors="1" failures="0" skip="0">
        <testcase classname="path_to_test_suite.TestSomething"
                  name="test_it" time="0">
            <error type="exceptions.TypeError" message="oops, wrong type">
            Traceback (most recent call last):
            ...
            TypeError: oops, wrong type
            </error>
        </testcase>
    </testsuite>

.. _Hudson: https://hudson.dev.java.net/

"""

import doctest
import os
import traceback
import re
import inspect
from time import time
from xml.sax import saxutils

# Invalid XML characters, control characters 0-31 sans \t, \n and \r
CONTROL_CHARACTERS = re.compile(r"[\000-\010\013\014\016-\037]")

def xml_safe(value):
    """Replaces invalid XML characters with '?'."""
    return CONTROL_CHARACTERS.sub('?', value)

def escape_cdata(cdata):
    """Escape a string for an XML CDATA section."""
    return xml_safe(cdata).replace(']]>', ']]>]]&gt;<![CDATA[')

def nice_classname(obj):
    """Returns a nice name for class object or class instance.

        >>> nice_classname(Exception()) # doctest: +ELLIPSIS
        '...Exception'
        >>> nice_classname(Exception)
        'exceptions.Exception'

    """
    if inspect.isclass(obj):
        cls_name = obj.__name__
    else:
        cls_name = obj.__class__.__name__
    mod = inspect.getmodule(obj)
    if mod:
        name = mod.__name__
        # jython
        if name.startswith('org.python.core.'):
            name = name[len('org.python.core.'):]
        return "%s.%s" % (name, cls_name)
    else:
        return cls_name

def exc_message(exc_info):
    """Return the exception's message."""
    exc = exc_info[1]
    if exc is None:
        # str exception
        result = exc_info[0]
    else:
        try:
            result = str(exc)
        except UnicodeEncodeError:
            try:
                result = unicode(exc)
            except UnicodeError:
                # Fallback to args as neither str nor
                # unicode(Exception(u'\xe6')) work in Python < 2.6
                result = exc.args[0]
    return xml_safe(result)

class Xunit(object):
    """This plugin provides test results in the standard XUnit XML format."""
    name = 'xunit'
    xunit_file = "nosetests.xml"
    score = 2000
    encoding = 'UTF-8'
    error_report_file = None

    def _timeTaken(self):
        if hasattr(self, '_timer'):
            taken = time() - self._timer
        else:
            # test died before it ran (probably error in setup())
            # or success/failure added before test started probably 
            # due to custom TestResult munging
            taken = 0.0
        return taken

    def _quoteattr(self, attr):
        """Escape an XML attribute. Value can be unicode."""
        attr = xml_safe(attr)
        if isinstance(attr, unicode):
            attr = attr.encode(self.encoding)
        return saxutils.quoteattr(attr)

    def __init__(self):
        """Configures the xunit plugin."""
        self.stats = {'errors': 0,
                      'failures': 0,
                      'passes': 0,
                      'skipped': 0
                      }
        self.errorlist = []
        self.error_report_file = open(self.xunit_file, 'w')

    def report(self):
        """Writes an Xunit-formatted XML file

        The file includes a report of test errors and failures.

        """
        self.stats['encoding'] = self.encoding
        self.stats['total'] = (self.stats['errors'] + self.stats['failures']
                               + self.stats['passes'] + self.stats['skipped'])
        self.error_report_file.write(
            '<?xml version="1.0" encoding="%(encoding)s"?>'
            '<testsuite name="nosetests" tests="%(total)d" '
            'errors="%(errors)d" failures="%(failures)d" '
            'skip="%(skipped)d">' % self.stats)
        self.error_report_file.write(''.join(self.errorlist))
        self.error_report_file.write('</testsuite>')
        self.error_report_file.close()

    def addError(self, test, err, capt=None):
        """Add error output to Xunit report.
        """
        taken = self._timeTaken()

        if issubclass(err[0], SkipTest):
            type = 'skipped'
            self.stats['skipped'] += 1
        else:
            type = 'error'
            self.stats['errors'] += 1
        tb = ''.join(traceback.format_exception(*err))
        id = test.id()
        self.errorlist.append(
            '<testcase classname=%(cls)s name=%(name)s time="%(taken)d">'
            '<%(type)s type=%(errtype)s message=%(message)s><![CDATA[%(tb)s]]>'
            '</%(type)s></testcase>' %
            {'cls': self._quoteattr('.'.join(id.split('.')[:-1])),
             'name': self._quoteattr(id.split('.')[-1]),
             'taken': taken,
             'type': type,
             'errtype': self._quoteattr(nice_classname(err[0])),
             'message': self._quoteattr(exc_message(err)),
             'tb': escape_cdata(tb),
             })

    def addFailure(self, test, err, capt=None, tb_info=None):
        """Add failure output to Xunit report.
        """
        taken = self._timeTaken()
        tb = ''.join(traceback.format_exception(*err))
        self.stats['failures'] += 1
        id = test
        self.errorlist.append(
            '<testcase classname=%(cls)s name=%(name)s time="%(taken)d">'
            '<failure type=%(errtype)s message=%(message)s><![CDATA[%(tb)s]]>'
            '</failure></testcase>' %
            {'cls': self._quoteattr('.'.join(id.split('.')[:-1])),
             'name': self._quoteattr(id.split('.')[-1]),
             'taken': taken,
             'errtype': self._quoteattr(nice_classname(err[0])),
             'message': self._quoteattr(exc_message(err)),
             'tb': escape_cdata(tb),
             })

    def addSuccess(self, test, capt=None):
        """Add success output to Xunit report.
        """
        taken = self._timeTaken()
        self.stats['passes'] += 1
        id = test
        self.errorlist.append(
            '<testcase classname=%(cls)s name=%(name)s '
            'time="%(taken)d" />' %
            {'cls': self._quoteattr('.'.join(id.split('.')[:-1])),
             'name': self._quoteattr(id.split('.')[-1]),
             'taken': taken,
             })

    def start(self):
        self._timer = time()



"""
Runner mechanism for answer testing

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import matplotlib
import os, shelve, cPickle, sys, imp, tempfile

from yt.config import ytcfg; ytcfg["yt","serialize"] = "False"
from yt.funcs import *
from yt.utilities.command_line import YTCommand
from .xunit import Xunit

from output_tests import test_registry, MultipleOutputTest, \
                         RegressionTestException

def clear_registry():
    test_registry.clear()

class FileNotExistException(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __repr__(self):
        return "FileNotExistException: %s" % (self.filename)


def registry_entries():
    return test_registry.keys()

class RegressionTestStorage(object):
    def __init__(self, results_id, path = "."):
        self.id = results_id
        if results_id == "":
            self._path = os.path.join(path, "results")
        else:
            self._path = os.path.join(path, "results_%s" % self.id)
        if not os.path.isdir(self._path): 
            only_on_root(os.mkdir, self._path)
        if os.path.isfile(self._path): raise RuntimeError

    def _fn(self, tn):
        return os.path.join(self._path, tn)

    @rootonly
    def __setitem__(self, test_name, result):
        # We have to close our shelf manually,
        # as the destructor does not necessarily do this.
        # Context managers would be more appropriate.
        f = open(self._fn(test_name), "wb")
        cPickle.dump(result, f, protocol=-1)
        f.close()

    def __getitem__(self, test_name):
        if not os.path.exists(self._fn(test_name)):
            raise FileNotExistException(self._fn(test_name))
        f = open(self._fn(test_name), "rb")
        tr = cPickle.load(f)
        f.close()
        return tr

class RegressionTestRunner(object):
    def __init__(self, results_id, compare_id = None,
                 results_path = ".", compare_results_path = ".",
                 io_log = "OutputLog", plot_tests = False):
        # This test runner assumes it has been launched with the current
        # working directory that of the test case itself.
        self.io_log = io_log
        self.id = results_id
        if compare_id is not None:
            self.old_results = RegressionTestStorage(
                                    compare_id, path=compare_results_path)
        else:
            self.old_results = None
        self.results = RegressionTestStorage(results_id, path=results_path)
        self.plot_list = {}
        self.passed_tests = {}
        self.test_messages = {}
        self.plot_tests = plot_tests

    def run_all_tests(self):
        plot_list = []
        for i,name in enumerate(sorted(test_registry)):
            self.run_test(name)
        return self.passed_tests

    def run_test(self, name):
        # We'll also need to call the "compare" operation,
        # but for that we'll need a data store.
        test = test_registry[name]
        plot_list = []
        if test.output_type == 'single':
            mot = MultipleOutputTest(self.io_log)
            for i,fn in enumerate(mot):
                # This next line is to keep the shelve module
                # from happily gobbling the disk
                #if i > 5: break 
                test_instance = test(fn)
                test_instance.name = "%s_%s" % (
                    os.path.basename(fn), test_instance.name )
                self._run(test_instance)

        elif test.output_type == 'multiple':
            test_instance = test(self.io_log)
            self._run(test_instance)

    watcher = None
    def _run(self, test):
        if self.watcher is not None:
            self.watcher.start()
        print self.id, "Running", test.name,
        test.setup()
        test.run()
        if self.plot_tests:
            self.plot_list[test.name] = test.plot()
        self.results[test.name] = test.result
        success, msg, exc = self._compare(test)
        if self.old_results is None:
            print "NO OLD RESULTS"
        else:
            if success == True: print "SUCCEEDED"
            else: print "FAILED", msg
        self.passed_tests[test.name] = success
        self.test_messages[test.name] = msg
        if self.watcher is not None:
            if success == True:
                self.watcher.addSuccess(test.name)
            else:
                self.watcher.addFailure(test.name, exc)

    def _compare(self, test):
        if self.old_results is None:
            return (True, "", "New Test")
        try:
            old_result = self.old_results[test.name]
        except FileNotExistException:
            return (False, sys.exc_info())
        try:
            test.compare(old_result)
        except RegressionTestException as exc:
            return (False, repr(exc), sys.exc_info())
        return (True, "", "Pass")

    def run_tests_from_file(self, filename):
        for line in open(filename):
            test_name = line.strip()
            if test_name not in test_registry:
                if test_name[0] != "#":
                    print "Test '%s' not recognized, skipping" % (test_name)
                continue
            print "Running '%s'" % (test_name)
            self.run_test(line.strip())

def _load_modules(test_modules):
    for fn in test_modules:
        if fn.endswith(".py"): fn = fn[:-3]
        print "Loading module %s" % (fn)
        mname = os.path.basename(fn)
        f, filename, desc = imp.find_module(mname, [os.path.dirname(fn)])
        project = imp.load_module(mname, f, filename, desc)

def _update_io_log(opts, kwargs):
    if opts.datasets is None or len(opts.datasets) == 0: return
    f = tempfile.NamedTemporaryFile()
    kwargs['io_log'] = f.name
    for d in opts.datasets:
        fn = os.path.expanduser(d)
        print "Registered dataset %s" % fn
        f.write("DATASET WRITTEN %s\n" % fn)
    f.flush()
    f.seek(0)
    return f



"""
Modifications and extensions to Bottle, to make it slightly more useful for
yt's purposes

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import uuid
import json
import logging, threading
import sys
import urllib, urllib2
import numpy as np

from yt.utilities.bottle import \
    server_names, debug, route, run, request, ServerAdapter, response
from extdirect_router import DirectRouter, DirectProviderDefinition
from yt.utilities.logger import ytLogger as mylog
from yt.funcs import *

route_functions = {}
route_watchers = []
payloads = []

def preroute(future_route, *args, **kwargs):
    def router(func):
        route_functions[future_route] = (args, kwargs, func)
        return func
    return router

def notify_route(watcher):
    route_watchers.append(watcher)

class BinaryDelivery(object):
    delivered = False
    payload = ""
    def __init__(self, payload, name = ""):
        self.name = name
        self.payload = payload
        #sys.__stderr__.write("CREATING A BINARY PAYLOAD %s (%s)\n" % (
        #    self.name, len(self.payload)))

    def get(self):
        # We set our 
        #sys.__stderr__.write("REQUESTED A BINARY PAYLOAD %s (%s)\n" % (
        #    self.name, len(self.payload)))
        p = self.payload
        if p == "":
            response.status = 404
            return
        self.payload = ""
        return p

class PayloadHandler(object):
    _shared_state = {}
    payloads = None
    binary_payloads = None
    recorded_payloads = None
    multicast_ids = None
    multicast_payloads = None
    lock = None
    record = False
    event = None
    count = 0
    debug = False
    _prefix = ""

    def __new__(cls, *p, **k):
        self = object.__new__(cls, *p, **k)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self):
        if self.payloads is None: self.payloads = []
        if self.lock is None: self.lock = threading.Lock()
        if self.recorded_payloads is None: self.recorded_payloads = []
        if self.event is None: self.event = threading.Event()
        if self.multicast_payloads is None: self.multicast_payloads = {}
        if self.multicast_ids is None: self.multicast_ids = {}
        if self.binary_payloads is None: self.binary_payloads = []

    def deliver_payloads(self):
        with self.lock:
            payloads = self.payloads
            if self.record:
                self.recorded_payloads += self.payloads
            if self.debug:
                sys.__stderr__.write("**** Delivering %s payloads\n" % (len(payloads)))
                for p in payloads:
                    sys.__stderr__.write("****    %s\n" % p['type'])
            self.payloads = []
            self.event.clear()
            try:
                self.deliver_multicast()
            except Exception as exc:
                sys.__stderr__.write("%s" % exc)
        return payloads

    def add_payload(self, to_add):
        with self.lock:
            if "binary" in to_add:
                self._add_binary_payload(to_add)
            self.payloads.append(to_add)
            # Does this next part need to be in the lock?
            if to_add.get("widget_id", None) in self.multicast_ids:
                self.multicast_payloads[to_add["widget_id"]] = to_add
            self.count += 1
            self.event.set()
            if self.debug:
                sys.__stderr__.write("**** Adding payload of type %s\n" % (to_add['type']))

    def _add_binary_payload(self, bp):  
        # This shouldn't be called by anybody other than add_payload.
        bkeys = ensure_list(bp['binary'])
        bp['binary'] = []
        for bkey in bkeys:
            bdata = bp.pop(bkey) # Get the binary data
            if isinstance(bdata, np.ndarray):
                bdata = bdata.tostring()
            bpserver = BinaryDelivery(bdata, bkey)
            self.binary_payloads.append(bpserver)
            uu = uuid.uuid4().hex
            bp['binary'].append((bkey, uu))
            route("%s/%s" % (self._prefix, uu))(bpserver.get)
            if self.debug:
                sys.__stderr__.write(
                    "**** Adding binary payload (%s) to %s\n" % (bkey, uu))

    def replay_payloads(self):
        return self.recorded_payloads

    def widget_payload(self, widget, data):
        data['type'] = 'widget_payload'
        data['widget_id'] = widget._ext_widget_id
        self.add_payload(data)

    def deliver_multicast(self):
        for widget_id in self.multicast_payloads:
            if widget_id not in self.multicast_payloads: continue
            server_id, session_token = self.multicast_ids[widget_id]
            # Now we execute a post to the correct location
            data = urllib.urlencode({
                'payload_session_id': server_id,
                'payload_session_token': session_token,
                'payload_data': self.multicast_payloads[widget_id],
                'payload_metadata': {}
            })
            urllib2.urlopen("http://localhost:8080/UpdatePayload", data = data)

class YTRocketServer(ServerAdapter):
    server_info = {} # Hack to get back at instance vars
    def run(self, handler):
        from yt.utilities.rocket import Rocket
        server = Rocket((self.host, self.port), 'wsgi', { 'wsgi_app' : handler })
        self.server_info[id(self)] = server
        server.start()

class BottleDirectRouter(DirectRouter):
    # This class implements a mechanism for auto-routing an ExtDirect-callable
    # object through Bottle.  It should be used as a base class of an object,
    # and the __init__ function will need to include the keyword argument
    # 'route' for it to work.
    _route_prefix = None
    def __init__(self, *args, **kwargs):
        future_route = self.api_url
        super(BottleDirectRouter, self).__init__(*args, **kwargs)
        self.__name__ = str(self.my_name)
        route_functions[future_route] = ((), {'method':("POST", "GET")}, self)
        preroute("/resources/ext-%s-api.js" % self.api_url, method="GET")(self._myapi)
        notify_route(self)

    def _myapi(self):
        dpd = DirectProviderDefinition(self, self.api_url, ns="yt_rpc")
        source = "Ext.Direct.addProvider(%s);" % json.dumps(dpd._config())
        response.headers['Content-Type'] = "text/javascript"
        return source

    def __call__(self):
        #print "Hi there, I just got this request:",
        val = request.body.read()
        #print val
        #import pdb;pdb.set_trace()
        rv = super(BottleDirectRouter, self).__call__(val)
        #print "With this response:", rv
        return rv

def uuid_serve_functions(pre_routed = None, open_browser=False, port=9099,
                         repl = None, token = None):
    if pre_routed == None: pre_routed = route_functions
    debug(mode=True)
    if token is None: token = uuid.uuid1()
    for r in pre_routed:
        args, kwargs, f = pre_routed[r]
        if r[0] == "/": r = r[1:]
        rp = "/%s/%s" % (token, r)
        func_name = getattr(f, 'func_name', str(f))
        print "Routing from %s => %s" % (rp, func_name)
        route(rp, *args, **kwargs)(f)
    for w in route_watchers:
        if not hasattr(w, "_route_prefix"):
            print "WARNING: %s has no _route_prefix attribute.  Not notifying."
            continue
            w._route_prefix = token
    repl._global_token = token
    repl.activate()
    repl.execution_thread.wait()
    print
    print
    print "============================================================================="
    print "============================================================================="
    print "Greetings, and welcome to Reason!"
    print "Your private token is %s ." % token
    print "DO NOT SHARE THIS TOKEN."
    print
    print "Please direct your browser to:"
    print
    print "     http://localhost:%s/%s/" % (port, token)
    print
    print "============================================================================="
    print
    print "If you are currently ssh'd into a remote machine, you should be able"
    print "to create a new SSH tunnel by typing or copy/pasting this text"
    print "verbatim, while waiting to see the 'ssh>' prompt after the first line."
    print
    print "~C"
    print "-L%s:localhost:%s" % (port, port)
    print
    print "and then pointing a web browser on your local machine to the above URL."
    print
    print "============================================================================="
    print "============================================================================="
    print
    print
    if open_browser:
        # We do some fancy footwork so that we can open the browser while the
        # server starts up.  I got this from some recipe whose URL escapes me.
        # Thank you, to whoever wrote it!
        def local_browse():
            """Start a browser after waiting for half a second."""
            import webbrowser, threading
            def _local_browse():
                webbrowser.open('http://localhost:%s/%s/' % (port, token))
            thread = threading.Timer(0.5, _local_browse)
            thread.start()
        local_browse()
    try:
        import yt.utilities.rocket as rocket
        server_type = YTRocketServer
        log = logging.getLogger('Rocket')
        log.setLevel(logging.WARNING)
        kwargs = {'timeout': 600, 'max_threads': 2}
        if repl is not None:
            repl.server = YTRocketServer.server_info
    except ImportError:
        server_type = server_names.get("wsgiref")
        kwargs = {}
    server = server_type(host='localhost', port=port, **kwargs)
    mylog.info("Starting up the server.")
    run(server=server)

class MethodLock(object):
    _shared_state = {}
    locks = None

    def __new__(cls, *p, **k):
        self = object.__new__(cls, *p, **k)
        self.__dict__ = cls._shared_state
        return self

    def __init__(self):
        if self.locks is None: self.locks = {}

    def __call__(self, func):
        if str(func) not in self.locks:
            self.locks[str(func)] = threading.Lock()
        @wraps(func)
        def locker(*args, **kwargs):
            print "Acquiring lock on %s" % (str(func))
            with self.locks[str(func)]:
                rv = func(*args, **kwargs)
            print "Regained lock on %s" % (str(func))
            return rv
        return locker

lockit = MethodLock()




"""
Field Interpolation Tables

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

cimport cython
cimport numpy as np
from fp_utils cimport imax, fmax, imin, fmin, iclip, fclip, fabs

DEF Nch = 4

cdef struct FieldInterpolationTable:
    # Note that we make an assumption about retaining a reference to values
    # externally.
    np.float64_t *values 
    np.float64_t bounds[2]
    np.float64_t dbin
    np.float64_t idbin
    int field_id
    int weight_field_id
    int weight_table_id
    int nbins

cdef extern from "math.h": 
    double expf(double x) nogil 

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void FIT_initialize_table(FieldInterpolationTable *fit, int nbins,
              np.float64_t *values, np.float64_t bounds1, np.float64_t bounds2,
              int field_id, int weight_field_id, int weight_table_id) nogil:
    fit.bounds[0] = bounds1; fit.bounds[1] = bounds2
    fit.nbins = nbins
    fit.dbin = (fit.bounds[1] - fit.bounds[0])/(fit.nbins-1)
    fit.idbin = 1.0/fit.dbin
    # Better not pull this out from under us, yo
    fit.values = values
    fit.field_id = field_id
    fit.weight_field_id = weight_field_id
    fit.weight_table_id = weight_table_id

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline np.float64_t FIT_get_value(FieldInterpolationTable *fit,
                                       np.float64_t dvs[6]) nogil:
    cdef np.float64_t bv, dy, dd, tf, rv
    cdef int bin_id
    if dvs[fit.field_id] >= fit.bounds[1] or dvs[fit.field_id] <= fit.bounds[0]: return 0.0
    bin_id = <int> ((dvs[fit.field_id] - fit.bounds[0]) * fit.idbin)
    bin_id = iclip(bin_id, 0, fit.nbins-2)
    dd = dvs[fit.field_id] - (fit.bounds[0] + bin_id * fit.dbin) # x - x0
    bv = fit.values[bin_id]
    dy = fit.values[bin_id + 1] - bv
    if fit.weight_field_id != -1:
        return dvs[fit.weight_field_id] * (bv + dd*dy*fit.idbin)
    return (bv + dd*dy*fit.idbin)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void FIT_eval_transfer(np.float64_t dt, np.float64_t *dvs,
                            np.float64_t *rgba, int n_fits,
                            FieldInterpolationTable fits[6],
                            int field_table_ids[6], int grey_opacity) nogil:
    cdef int i, fid, use
    cdef np.float64_t ta, tf, ttot, istorage[6], trgba[6], dot_prod
    for i in range(6): istorage[i] = 0.0
    for i in range(n_fits):
        istorage[i] = FIT_get_value(&fits[i], dvs)
    for i in range(n_fits):
        fid = fits[i].weight_table_id
        if fid != -1: istorage[i] *= istorage[fid]
    for i in range(6):
        trgba[i] = istorage[field_table_ids[i]]

    if grey_opacity == 1:
        ta = fmax(1.0 - dt*trgba[3],0.0)
        for i in range(4):
            rgba[i] = dt*trgba[i] + ta*rgba[i]
    else:
        for i in range(3):
            ta = fmax(1.0-dt*trgba[i], 0.0)
            rgba[i] = dt*trgba[i] + ta*rgba[i]

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void FIT_eval_transfer_with_light(np.float64_t dt, np.float64_t *dvs, 
        np.float64_t *grad, np.float64_t *l_dir, np.float64_t *l_rgba,
        np.float64_t *rgba, int n_fits,
        FieldInterpolationTable fits[6],
        int field_table_ids[6], int grey_opacity) nogil:
    cdef int i, fid, use
    cdef np.float64_t ta, tf, istorage[6], trgba[6], dot_prod
    dot_prod = 0.0
    for i in range(3):
        dot_prod += l_dir[i]*grad[i]
    #dot_prod = fmax(0.0, dot_prod)
    for i in range(6): istorage[i] = 0.0
    for i in range(n_fits):
        istorage[i] = FIT_get_value(&fits[i], dvs)
    for i in range(n_fits):
        fid = fits[i].weight_table_id
        if fid != -1: istorage[i] *= istorage[fid]
    for i in range(6):
        trgba[i] = istorage[field_table_ids[i]]
    if grey_opacity == 1:
        ta = fmax(1.0-dt*(trgba[0] + trgba[1] + trgba[2]), 0.0)
        for i in range(3):
            rgba[i] = (1.-ta)*trgba[i]*(1. + dot_prod*l_rgba[i]) + ta * rgba[i]
    else:
        for i in range(3):
            ta = fmax(1.0-dt*trgba[i], 0.0)
            rgba[i] = (1.-ta)*trgba[i]*(1. + dot_prod*l_rgba[i]) + ta * rgba[i]




"""
The basic field info container resides here.  These classes, code specific and
universal, are the means by which we access fields across YT, both derived and
native.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import numpy as np
import inspect
import copy

from yt.funcs import *

from yt.utilities.lib import CICDeposit_3, obtain_rvec, obtain_rv_vec
from yt.utilities.cosmology import Cosmology
from field_info_container import \
    add_field, \
    ValidateDataField, \
    ValidateGridType, \
    ValidateParameter, \
    ValidateSpatial, \
    NeedsGridType, \
    NeedsOriginalGrid, \
    NeedsDataField, \
    NeedsProperty, \
    NeedsParameter, \
    NullFunc

from yt.utilities.physical_constants import \
     mh, \
     me, \
     sigma_thompson, \
     clight, \
     kboltz, \
     G, \
     rho_crit_now, \
     speed_of_light_cgs, \
     km_per_cm, keV_per_K

from yt.utilities.math_utils import \
    get_sph_r_component, \
    get_sph_theta_component, \
    get_sph_phi_component, \
    get_cyl_r_component, \
    get_cyl_z_component, \
    get_cyl_theta_component, \
    get_cyl_r, get_cyl_theta, \
    get_cyl_z, get_sph_r, \
    get_sph_theta, get_sph_phi, \
    periodic_dist, euclidean_dist
     
# Note that, despite my newfound efforts to comply with PEP-8,
# I violate it here in order to keep the name/func_name relationship

def _dx(field, data):
    return data.dds[0]
    return np.ones(data.ActiveDimensions, dtype='float64') * data.dds[0]
add_field('dx', function=_dx, display_field=False,
          validators=[ValidateSpatial(0)])

def _dy(field, data):
    return data.dds[1]
    return np.ones(data.ActiveDimensions, dtype='float64') * data.dds[1]
add_field('dy', function=_dy, display_field=False,
          validators=[ValidateSpatial(0)])

def _dz(field, data):
    return data.dds[2]
    return np.ones(data.ActiveDimensions, dtype='float64') * data.dds[2]
add_field('dz', function=_dz,
          display_field=False, validators=[ValidateSpatial(0)])

def _coordX(field, data):
    dim = data.ActiveDimensions[0]
    return (np.ones(data.ActiveDimensions, dtype='float64')
                   * np.arange(data.ActiveDimensions[0])[:,None,None]
            +0.5) * data['dx'] + data.LeftEdge[0]
add_field('x', function=_coordX, display_field=False,
          validators=[ValidateSpatial(0)])

def _coordY(field, data):
    dim = data.ActiveDimensions[1]
    return (np.ones(data.ActiveDimensions, dtype='float64')
                   * np.arange(data.ActiveDimensions[1])[None,:,None]
            +0.5) * data['dy'] + data.LeftEdge[1]
add_field('y', function=_coordY, display_field=False,
          validators=[ValidateSpatial(0)])

def _coordZ(field, data):
    dim = data.ActiveDimensions[2]
    return (np.ones(data.ActiveDimensions, dtype='float64')
                   * np.arange(data.ActiveDimensions[2])[None,None,:]
            +0.5) * data['dz'] + data.LeftEdge[2]
add_field('z', function=_coordZ, display_field=False,
          validators=[ValidateSpatial(0)])

def _GridLevel(field, data):
    return np.ones(data.ActiveDimensions)*(data.Level)
add_field("GridLevel", function=_GridLevel,
          validators=[ValidateGridType(),
                      ValidateSpatial(0)])

def _GridIndices(field, data):
    return np.ones(data["Ones"].shape)*(data.id-data._id_offset)
add_field("GridIndices", function=_GridIndices,
          validators=[ValidateGridType(),
                      ValidateSpatial(0)], take_log=False)

def _OnesOverDx(field, data):
    return np.ones(data["Ones"].shape,
                   dtype=data["Density"].dtype)/data['dx']
add_field("OnesOverDx", function=_OnesOverDx,
          display_field=False)

def _Zeros(field, data):
    return np.zeros(data.ActiveDimensions, dtype='float64')
add_field("Zeros", function=_Zeros,
          validators=[ValidateSpatial(0)],
          projection_conversion="unitary",
          display_field = False)

def _Ones(field, data):
    return np.ones(data.ActiveDimensions, dtype='float64')
add_field("Ones", function=_Ones,
          validators=[ValidateSpatial(0)],
          projection_conversion="unitary",
          display_field = False)
add_field("CellsPerBin", function=_Ones, validators=[ValidateSpatial(0)],
          display_field = False)

def _SoundSpeed(field, data):
    if data.pf["EOSType"] == 1:
        return np.ones(data["Density"].shape, dtype='float64') * \
                data.pf["EOSSoundSpeed"]
    return ( data.pf["Gamma"]*data["Pressure"] / \
             data["Density"] )**(1.0/2.0)
add_field("SoundSpeed", function=_SoundSpeed,
          units=r"\rm{cm}/\rm{s}")

def _RadialMachNumber(field, data):
    """M{|v|/t_sound}"""
    return np.abs(data["RadialVelocity"]) / data["SoundSpeed"]
add_field("RadialMachNumber", function=_RadialMachNumber)

def _MachNumber(field, data):
    """M{|v|/t_sound}"""
    return data["VelocityMagnitude"] / data["SoundSpeed"]
add_field("MachNumber", function=_MachNumber)

def _CourantTimeStep(field, data):
    t1 = data['dx'] / (
        data["SoundSpeed"] + \
        abs(data["x-velocity"]))
    t2 = data['dy'] / (
        data["SoundSpeed"] + \
        abs(data["y-velocity"]))
    t3 = data['dz'] / (
        data["SoundSpeed"] + \
        abs(data["z-velocity"]))
    return np.minimum(np.minimum(t1,t2),t3)
def _convertCourantTimeStep(data):
    # SoundSpeed and z-velocity are in cm/s, dx is in code
    return data.convert("cm")
add_field("CourantTimeStep", function=_CourantTimeStep,
          convert_function=_convertCourantTimeStep,
          units=r"$\rm{s}$")

def _ParticleVelocityMagnitude(field, data):
    """M{|v|}"""
    bulk_velocity = data.get_field_parameter("bulk_velocity")
    if bulk_velocity == None:
        bulk_velocity = np.zeros(3)
    return ( (data["particle_velocity_x"]-bulk_velocity[0])**2.0 + \
             (data["particle_velocity_y"]-bulk_velocity[1])**2.0 + \
             (data["particle_velocity_z"]-bulk_velocity[2])**2.0 )**(1.0/2.0)
add_field("ParticleVelocityMagnitude", function=_ParticleVelocityMagnitude,
          particle_type=True, 
          take_log=False, units=r"\rm{cm}/\rm{s}")

def _VelocityMagnitude(field, data):
    """M{|v|}"""
    velocities = obtain_rv_vec(data)
    return np.sqrt(np.sum(velocities**2,axis=0))
add_field("VelocityMagnitude", function=_VelocityMagnitude,
          take_log=False, units=r"\rm{cm}/\rm{s}")

def _TangentialOverVelocityMagnitude(field, data):
    return np.abs(data["TangentialVelocity"])/np.abs(data["VelocityMagnitude"])
add_field("TangentialOverVelocityMagnitude",
          function=_TangentialOverVelocityMagnitude,
          take_log=False)

def _Pressure(field, data):
    """M{(Gamma-1.0)*rho*E}"""
    return (data.pf["Gamma"] - 1.0) * \
           data["Density"] * data["ThermalEnergy"]
add_field("Pressure", function=_Pressure, units=r"\rm{dyne}/\rm{cm}^{2}")

def _TempkeV(field, data):
    return data["Temperature"] * keV_per_K
add_field("TempkeV", function=_TempkeV, units=r"\rm{keV}",
          display_name="Temperature")

def _Entropy(field, data):
    if data.has_field_parameter("mu"):
        mw = mh*data.get_field_parameter("mu")
    else :
        mw = mh
    try:
        gammam1 = data.pf["Gamma"] - 1.0
    except:
        gammam1 = 5./3. - 1.0
    return kboltz * data["Temperature"] / \
           ((data["Density"]/mw)**gammam1)
add_field("Entropy", units=r"\rm{ergs}\ \rm{cm}^{3\gamma-3}",
          function=_Entropy)

### spherical coordinates: r (radius)
def _sph_r(field, data):
    center = data.get_field_parameter("center")
      
    coords = obtain_rvec(data)

    return get_sph_r(coords)

def _Convert_sph_r_CGS(data):
   return data.convert("cm")

add_field("sph_r", function=_sph_r,
         validators=[ValidateParameter("center")],
         convert_function = _Convert_sph_r_CGS, units=r"\rm{cm}")


### spherical coordinates: theta (angle with respect to normal)
def _sph_theta(field, data):
    center = data.get_field_parameter("center")
    normal = data.get_field_parameter("normal")
    
    coords = obtain_rvec(data)

    return get_sph_theta(coords, normal)

add_field("sph_theta", function=_sph_theta,
         validators=[ValidateParameter("center"),ValidateParameter("normal")])


### spherical coordinates: phi (angle in the plane perpendicular to the normal)
def _sph_phi(field, data):
    center = data.get_field_parameter("center")
    normal = data.get_field_parameter("normal")
    
    coords = obtain_rvec(data)

    return get_sph_phi(coords, normal)

add_field("sph_phi", function=_sph_phi,
         validators=[ValidateParameter("center"),ValidateParameter("normal")])

### cylindrical coordinates: R (radius in the cylinder's plane)
def _cyl_R(field, data):
    center = data.get_field_parameter("center")
    normal = data.get_field_parameter("normal")
      
    coords = obtain_rvec(data)

    return get_cyl_r(coords, normal)

def _Convert_cyl_R_CGS(data):
   return data.convert("cm")

add_field("cyl_R", function=_cyl_R,
         validators=[ValidateParameter("center"),ValidateParameter("normal")],
         convert_function = _Convert_cyl_R_CGS, units=r"\rm{cm}")
add_field("cyl_RCode", function=_cyl_R,
          validators=[ValidateParameter("center"),ValidateParameter("normal")],
          units=r"Radius (code)")


### cylindrical coordinates: z (height above the cylinder's plane)
def _cyl_z(field, data):
    center = data.get_field_parameter("center")
    normal = data.get_field_parameter("normal")
    
    coords = obtain_rvec(data)

    return get_cyl_z(coords, normal)

def _Convert_cyl_z_CGS(data):
   return data.convert("cm")

add_field("cyl_z", function=_cyl_z,
         validators=[ValidateParameter("center"),ValidateParameter("normal")],
         convert_function = _Convert_cyl_z_CGS, units=r"\rm{cm}")


### cylindrical coordinates: theta (angle in the cylinder's plane)
def _cyl_theta(field, data):
    center = data.get_field_parameter("center")
    normal = data.get_field_parameter("normal")
    
    coords = obtain_rvec(data)

    return get_cyl_theta(coords, normal)

add_field("cyl_theta", function=_cyl_theta,
         validators=[ValidateParameter("center"),ValidateParameter("normal")])

### The old field DiskAngle is the same as the spherical coordinates'
### 'theta' angle. I'm keeping DiskAngle for backwards compatibility.
def _DiskAngle(field, data):
    return data['sph_theta']

add_field("DiskAngle", function=_DiskAngle,
          take_log=False,
          validators=[ValidateParameter("center"),
                      ValidateParameter("normal")],
          display_field=False)


### The old field Height is the same as the cylindrical coordinates' z
### field. I'm keeping Height for backwards compatibility.
def _Height(field, data):
    return data['cyl_z']

def _convertHeight(data):
    return data.convert("cm")
def _convertHeightAU(data):
    return data.convert("au")
add_field("Height", function=_Height,
          convert_function=_convertHeight,
          validators=[ValidateParameter("center"),
                      ValidateParameter("normal")],
          units=r"cm", display_field=False)
add_field("HeightAU", function=_Height,
          convert_function=_convertHeightAU,
          validators=[ValidateParameter("center"),
                      ValidateParameter("normal")],
          units=r"AU", display_field=False)

def _cyl_RadialVelocity(field, data):
    normal = data.get_field_parameter("normal")
    velocities = obtain_rv_vec(data)

    theta = data['cyl_theta']

    return get_cyl_r_component(velocities, theta, normal)

def _cyl_RadialVelocityABS(field, data):
    return np.abs(_cyl_RadialVelocity(field, data))
def _Convert_cyl_RadialVelocityKMS(data):
    return km_per_cm
add_field("cyl_RadialVelocity", function=_cyl_RadialVelocity,
          units=r"\rm{cm}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_RadialVelocityABS", function=_cyl_RadialVelocityABS,
          units=r"\rm{cm}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_RadialVelocityKMS", function=_cyl_RadialVelocity,
          convert_function=_Convert_cyl_RadialVelocityKMS, units=r"\rm{km}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_RadialVelocityKMSABS", function=_cyl_RadialVelocityABS,
          convert_function=_Convert_cyl_RadialVelocityKMS, units=r"\rm{km}/\rm{s}",
          validators=[ValidateParameter("normal")])

def _cyl_TangentialVelocity(field, data):
    normal = data.get_field_parameter("normal")
    velocities = obtain_rv_vec(data)
    theta = data['cyl_theta']

    return get_cyl_theta_component(velocities, theta, normal)

def _cyl_TangentialVelocityABS(field, data):
    return np.abs(_cyl_TangentialVelocity(field, data))
def _Convert_cyl_TangentialVelocityKMS(data):
    return km_per_cm
add_field("cyl_TangentialVelocity", function=_cyl_TangentialVelocity,
          units=r"\rm{cm}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_TangentialVelocityABS", function=_cyl_TangentialVelocityABS,
          units=r"\rm{cm}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_TangentialVelocityKMS", function=_cyl_TangentialVelocity,
          convert_function=_Convert_cyl_TangentialVelocityKMS, units=r"\rm{km}/\rm{s}",
          validators=[ValidateParameter("normal")])
add_field("cyl_TangentialVelocityKMSABS", function=_cyl_TangentialVelocityABS,
          convert_function=_Convert_cyl_TangentialVelocityKMS, units=r"\rm{km}/\rm{s}",
          validators=[ValidateParameter("normal")])

def _DynamicalTime(field, data):
    """
    The formulation for the dynamical time is:
    M{sqrt(3pi/(16*G*rho))} or M{sqrt(3pi/(16G))*rho^-(1/2)}
    Note that we return in our natural units already
    """
    return (3.0*np.pi/(16*G*data["Density"]))**(1./2.)
add_field("DynamicalTime", function=_DynamicalTime,
           units=r"\rm{s}")

def JeansMassMsun(field,data):
    return (MJ_constant * 
            ((data["Temperature"]/data["MeanMolecularWeight"])**(1.5)) *
            (data["Density"]**(-0.5)))
add_field("JeansMassMsun",function=JeansMassMsun,units=r"\rm{Msun}")

def _CellMass(field, data):
    return data["Density"] * data["CellVolume"]
def _convertCellMassMsun(data):
    return 5.027854e-34 # g^-1
add_field("CellMass", function=_CellMass, units=r"\rm{g}")
add_field("CellMassMsun", units=r"M_{\odot}",
          function=_CellMass,
          convert_function=_convertCellMassMsun)

def _CellMassCode(field, data):
    return data["Density"] * data["CellVolumeCode"]
def _convertCellMassCode(data):
    return 1.0/data.convert("Density")
add_field("CellMassCode", 
          function=_CellMassCode,
          convert_function=_convertCellMassCode)

def _TotalMass(field,data):
    return (data["Density"]+data["particle_density"]) * data["CellVolume"]
add_field("TotalMass", function=_TotalMass, units=r"\rm{g}")
add_field("TotalMassMsun", units=r"M_{\odot}",
          function=_TotalMass,
          convert_function=_convertCellMassMsun)

def _StarMass(field,data):
    return data["star_density"] * data["CellVolume"]
add_field("StarMassMsun", units=r"M_{\odot}",
          function=_StarMass,
          convert_function=_convertCellMassMsun)

def _Matter_Density(field,data):
    return (data['Density'] + data['particle_density'])
add_field("Matter_Density",function=_Matter_Density,units=r"\rm{g}/\rm{cm^3}")

def _ComovingDensity(field, data):
    ef = (1.0 + data.pf.current_redshift)**3.0
    return data["Density"]/ef
add_field("ComovingDensity", function=_ComovingDensity, units=r"\rm{g}/\rm{cm}^3")

# This is rho_total / rho_cr(z).
def _Convert_Overdensity(data):
    return 1.0 / (rho_crit_now * data.pf.hubble_constant**2 * 
                (1+data.pf.current_redshift)**3)
add_field("Overdensity",function=_Matter_Density,
          convert_function=_Convert_Overdensity, units=r"")

# This is (rho_total - <rho_total>) / <rho_total>.
def _DensityPerturbation(field, data):
    rho_bar = rho_crit_now * data.pf.omega_matter * \
        data.pf.hubble_constant**2 * \
        (1.0 + data.pf.current_redshift)**3
    return ((data['Matter_Density'] - rho_bar) / rho_bar)
add_field("DensityPerturbation",function=_DensityPerturbation,units=r"")

# This is rho_b / <rho_b>.
def _Baryon_Overdensity(field, data):
    if data.pf.has_key('omega_baryon_now'):
        omega_baryon_now = data.pf['omega_baryon_now']
    else:
        omega_baryon_now = 0.0441
    return data['Density'] / (omega_baryon_now * rho_crit_now * 
                              (data.pf.hubble_constant**2) * 
                              ((1+data.pf.current_redshift)**3))
add_field("Baryon_Overdensity", function=_Baryon_Overdensity, 
          units=r"")

# Weak lensing convergence.
# Eqn 4 of Metzler, White, & Loken (2001, ApJ, 547, 560).
def _convertConvergence(data):
    if not data.pf.parameters.has_key('cosmology_calculator'):
        data.pf.parameters['cosmology_calculator'] = Cosmology(
            HubbleConstantNow=(100.*data.pf.hubble_constant),
            OmegaMatterNow=data.pf.omega_matter, OmegaLambdaNow=data.pf.omega_lambda)
    # observer to lens
    DL = data.pf.parameters['cosmology_calculator'].AngularDiameterDistance(
        data.pf.parameters['observer_redshift'], data.pf.current_redshift)
    # observer to source
    DS = data.pf.parameters['cosmology_calculator'].AngularDiameterDistance(
        data.pf.parameters['observer_redshift'], data.pf.parameters['lensing_source_redshift'])
    # lens to source
    DLS = data.pf.parameters['cosmology_calculator'].AngularDiameterDistance(
        data.pf.current_redshift, data.pf.parameters['lensing_source_redshift'])
    return (((DL * DLS) / DS) * (1.5e14 * data.pf.omega_matter * 
                                (data.pf.hubble_constant / speed_of_light_cgs)**2 *
                                (1 + data.pf.current_redshift)))
add_field("WeakLensingConvergence", function=_DensityPerturbation, 
          convert_function=_convertConvergence, 
          projection_conversion='mpccm')

def _CellVolume(field, data):
    if data['dx'].size == 1:
        try:
            return data['dx'] * data['dy'] * data['dz'] * \
                np.ones(data.ActiveDimensions, dtype='float64')
        except AttributeError:
            return data['dx'] * data['dy'] * data['dz']
    return data["dx"] * data["dy"] * data["dz"]
def _ConvertCellVolumeMpc(data):
    return data.convert("mpc")**3.0
def _ConvertCellVolumeCGS(data):
    return data.convert("cm")**3.0
add_field("CellVolumeCode", units=r"\rm{BoxVolume}^3",
          function=_CellVolume)
add_field("CellVolumeMpc", units=r"\rm{Mpc}^3",
          function=_CellVolume,
          convert_function=_ConvertCellVolumeMpc)
add_field("CellVolume", units=r"\rm{cm}^3",
          function=_CellVolume,
          convert_function=_ConvertCellVolumeCGS)

def _ChandraEmissivity(field, data):
    logT0 = np.log10(data["Temperature"]) - 7
    return ((data["NumberDensity"].astype('float64')**2.0) \
            *(10**(-0.0103*logT0**8 \
                   +0.0417*logT0**7 \
                   -0.0636*logT0**6 \
                   +0.1149*logT0**5 \
                   -0.3151*logT0**4 \
                   +0.6655*logT0**3 \
                   -1.1256*logT0**2 \
                   +1.0026*logT0**1 \
                   -0.6984*logT0) \
              +data["Metallicity"]*10**(0.0305*logT0**11 \
                                        -0.0045*logT0**10 \
                                        -0.3620*logT0**9 \
                                        +0.0513*logT0**8 \
                                        +1.6669*logT0**7 \
                                        -0.3854*logT0**6 \
                                        -3.3604*logT0**5 \
                                        +0.4728*logT0**4 \
                                        +4.5774*logT0**3 \
                                        -2.3661*logT0**2 \
                                        -1.6667*logT0**1 \
                                        -0.2193*logT0)))
def _convertChandraEmissivity(data):
    return 1.0 #1.0e-23*0.76**2
add_field("ChandraEmissivity", function=_ChandraEmissivity,
          convert_function=_convertChandraEmissivity,
          projection_conversion="1")

def _XRayEmissivity(field, data):
    return ((data["Density"].astype('float64')**2.0) \
            *data["Temperature"]**0.5)
def _convertXRayEmissivity(data):
    return 2.168e60
add_field("XRayEmissivity", function=_XRayEmissivity,
          convert_function=_convertXRayEmissivity,
          projection_conversion="1")

def _SZKinetic(field, data):
    vel_axis = data.get_field_parameter('axis')
    if vel_axis > 2:
        raise NeedsParameter(['axis'])
    vel = data["%s-velocity" % ({0:'x',1:'y',2:'z'}[vel_axis])]
    return (vel*data["Density"])
def _convertSZKinetic(data):
    return 0.88*((sigma_thompson/mh)/clight)
add_field("SZKinetic", function=_SZKinetic,
          convert_function=_convertSZKinetic,
          validators=[ValidateParameter('axis')])

def _SZY(field, data):
    return (data["Density"]*data["Temperature"])
def _convertSZY(data):
    conv = (0.88/mh) * (kboltz)/(me * clight*clight) * sigma_thompson
    return conv
add_field("SZY", function=_SZY, convert_function=_convertSZY)

def _AveragedDensity(field, data):
    nx, ny, nz = data["Density"].shape
    new_field = np.zeros((nx-2,ny-2,nz-2), dtype='float64')
    weight_field = np.zeros((nx-2,ny-2,nz-2), dtype='float64')
    i_i, j_i, k_i = np.mgrid[0:3,0:3,0:3]
    for i,j,k in zip(i_i.ravel(),j_i.ravel(),k_i.ravel()):
        sl = [slice(i,nx-(2-i)),slice(j,ny-(2-j)),slice(k,nz-(2-k))]
        new_field += data["Density"][sl] * data["CellMass"][sl]
        weight_field += data["CellMass"][sl]
    # Now some fancy footwork
    new_field2 = np.zeros((nx,ny,nz))
    new_field2[1:-1,1:-1,1:-1] = new_field/weight_field
    return new_field2
add_field("AveragedDensity",
          function=_AveragedDensity,
          validators=[ValidateSpatial(1, ["Density"])])

def _DivV(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    ds = div_fac * data['dx'].flat[0]
    f  = data["x-velocity"][sl_right,1:-1,1:-1]/ds
    f -= data["x-velocity"][sl_left ,1:-1,1:-1]/ds
    if data.pf.dimensionality > 1:
        ds = div_fac * data['dy'].flat[0]
        f += data["y-velocity"][1:-1,sl_right,1:-1]/ds
        f -= data["y-velocity"][1:-1,sl_left ,1:-1]/ds
    if data.pf.dimensionality > 2:
        ds = div_fac * data['dz'].flat[0]
        f += data["z-velocity"][1:-1,1:-1,sl_right]/ds
        f -= data["z-velocity"][1:-1,1:-1,sl_left ]/ds
    new_field = np.zeros(data["x-velocity"].shape, dtype='float64')
    new_field[1:-1,1:-1,1:-1] = f
    return new_field
def _convertDivV(data):
    return data.convert("cm")**-1.0
add_field("DivV", function=_DivV,
            validators=[ValidateSpatial(1,
            ["x-velocity","y-velocity","z-velocity"])],
          units=r"\rm{s}^{-1}", take_log=False,
          convert_function=_convertDivV)

def _AbsDivV(field, data):
    return np.abs(data['DivV'])
add_field("AbsDivV", function=_AbsDivV,
          units=r"\rm{s}^{-1}")

def _Contours(field, data):
    return -np.ones_like(data["Ones"])
add_field("Contours", validators=[ValidateSpatial(0)], take_log=False,
          display_field=False, function=_Contours)
add_field("tempContours", function=_Contours,
          validators=[ValidateSpatial(0), ValidateGridType()],
          take_log=False, display_field=False)

def obtain_velocities(data):
    return obtain_rv_vec(data)

def _convertSpecificAngularMomentum(data):
    return data.convert("cm")
def _convertSpecificAngularMomentumKMSMPC(data):
    return data.convert("mpc")/1e5

def _SpecificAngularMomentumX(field, data):
    xv, yv, zv = obtain_velocities(data)
    rv = obtain_rvec(data)
    return yv*rv[2,:] - zv*rv[1,:]
def _SpecificAngularMomentumY(field, data):
    xv, yv, zv = obtain_velocities(data)
    rv = obtain_rvec(data)
    return -(xv*rv[2,:] - zv*rv[0,:])
def _SpecificAngularMomentumZ(field, data):
    xv, yv, zv = obtain_velocities(data)
    rv = obtain_rvec(data)
    return xv*rv[1,:] - yv*rv[0,:]
for ax in 'XYZ':
    n = "SpecificAngularMomentum%s" % ax
    add_field(n, function=eval("_%s" % n),
              convert_function=_convertSpecificAngularMomentum,
              units=r"\rm{cm}^2/\rm{s}", validators=[ValidateParameter("center")])

def _AngularMomentumX(field, data):
    return data["CellMass"] * data["SpecificAngularMomentumX"]
add_field("AngularMomentumX", function=_AngularMomentumX,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", vector_field=False,
         validators=[ValidateParameter('center')])
def _AngularMomentumY(field, data):
    return data["CellMass"] * data["SpecificAngularMomentumY"]
add_field("AngularMomentumY", function=_AngularMomentumY,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", vector_field=False,
         validators=[ValidateParameter('center')])
def _AngularMomentumZ(field, data):
    return data["CellMass"] * data["SpecificAngularMomentumZ"]
add_field("AngularMomentumZ", function=_AngularMomentumZ,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", vector_field=False,
         validators=[ValidateParameter('center')])

def _ParticleSpecificAngularMomentum(field, data):
    """
    Calculate the angular of a particle velocity.  Returns a vector for each
    particle.
    """
    if data.has_field_parameter("bulk_velocity"):
        bv = data.get_field_parameter("bulk_velocity")
    else: bv = np.zeros(3, dtype='float64')
    xv = data["particle_velocity_x"] - bv[0]
    yv = data["particle_velocity_y"] - bv[1]
    zv = data["particle_velocity_z"] - bv[2]
    center = data.get_field_parameter('center')
    coords = np.array([data['particle_position_x'],
                       data['particle_position_y'],
                       data['particle_position_z']], dtype='float64')
    new_shape = tuple([3] + [1]*(len(coords.shape)-1))
    r_vec = coords - np.reshape(center,new_shape)
    v_vec = np.array([xv,yv,zv], dtype='float64')
    return np.cross(r_vec, v_vec, axis=0)
#add_field("ParticleSpecificAngularMomentum",
#          function=_ParticleSpecificAngularMomentum, particle_type=True,
#          convert_function=_convertSpecificAngularMomentum, vector_field=True,
#          units=r"\rm{cm}^2/\rm{s}", validators=[ValidateParameter('center')])
def _convertSpecificAngularMomentumKMSMPC(data):
    return km_per_cm*data.convert("mpc")
#add_field("ParticleSpecificAngularMomentumKMSMPC",
#          function=_ParticleSpecificAngularMomentum, particle_type=True,
#          convert_function=_convertSpecificAngularMomentumKMSMPC, vector_field=True,
#          units=r"\rm{km}\rm{Mpc}/\rm{s}", validators=[ValidateParameter('center')])

def _ParticleSpecificAngularMomentumX(field, data):
    if data.has_field_parameter("bulk_velocity"):
        bv = data.get_field_parameter("bulk_velocity")
    else: bv = np.zeros(3, dtype='float64')
    center = data.get_field_parameter('center')
    y = data["particle_position_y"] - center[1]
    z = data["particle_position_z"] - center[2]
    yv = data["particle_velocity_y"] - bv[1]
    zv = data["particle_velocity_z"] - bv[2]
    return yv*z - zv*y
def _ParticleSpecificAngularMomentumY(field, data):
    if data.has_field_parameter("bulk_velocity"):
        bv = data.get_field_parameter("bulk_velocity")
    else: bv = np.zeros(3, dtype='float64')
    center = data.get_field_parameter('center')
    x = data["particle_position_x"] - center[0]
    z = data["particle_position_z"] - center[2]
    xv = data["particle_velocity_x"] - bv[0]
    zv = data["particle_velocity_z"] - bv[2]
    return -(xv*z - zv*x)
def _ParticleSpecificAngularMomentumZ(field, data):
    if data.has_field_parameter("bulk_velocity"):
        bv = data.get_field_parameter("bulk_velocity")
    else: bv = np.zeros(3, dtype='float64')
    center = data.get_field_parameter('center')
    x = data["particle_position_x"] - center[0]
    y = data["particle_position_y"] - center[1]
    xv = data["particle_velocity_x"] - bv[0]
    yv = data["particle_velocity_y"] - bv[1]
    return xv*y - yv*x
for ax in 'XYZ':
    n = "ParticleSpecificAngularMomentum%s" % ax
    add_field(n, function=eval("_%s" % n), particle_type=True,
              convert_function=_convertSpecificAngularMomentum,
              units=r"\rm{cm}^2/\rm{s}", validators=[ValidateParameter("center")])
    add_field(n + "KMSMPC", function=eval("_%s" % n), particle_type=True,
              convert_function=_convertSpecificAngularMomentumKMSMPC,
              units=r"\rm{cm}^2/\rm{s}", validators=[ValidateParameter("center")])

def _ParticleAngularMomentum(field, data):
    return data["ParticleMass"] * data["ParticleSpecificAngularMomentum"]
#add_field("ParticleAngularMomentum",
#          function=_ParticleAngularMomentum, units=r"\rm{g}\/\rm{cm}^2/\rm{s}",
#          particle_type=True, validators=[ValidateParameter('center')])
def _ParticleAngularMomentumMSUNKMSMPC(field, data):
    return data["ParticleMass"] * data["ParticleSpecificAngularMomentumKMSMPC"]
#add_field("ParticleAngularMomentumMSUNKMSMPC",
#          function=_ParticleAngularMomentumMSUNKMSMPC,
#          units=r"M_{\odot}\rm{km}\rm{Mpc}/\rm{s}",
#          particle_type=True, validators=[ValidateParameter('center')])

def _ParticleAngularMomentumX(field, data):
    return data["CellMass"] * data["ParticleSpecificAngularMomentumX"]
add_field("ParticleAngularMomentumX", function=_ParticleAngularMomentumX,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", particle_type=True,
         validators=[ValidateParameter('center')])
def _ParticleAngularMomentumY(field, data):
    return data["CellMass"] * data["ParticleSpecificAngularMomentumY"]
add_field("ParticleAngularMomentumY", function=_ParticleAngularMomentumY,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", particle_type=True,
         validators=[ValidateParameter('center')])
def _ParticleAngularMomentumZ(field, data):
    return data["CellMass"] * data["ParticleSpecificAngularMomentumZ"]
add_field("ParticleAngularMomentumZ", function=_ParticleAngularMomentumZ,
         units=r"\rm{g}\/\rm{cm}^2/\rm{s}", particle_type=True,
         validators=[ValidateParameter('center')])

def get_radius(data, field_prefix):
    center = data.get_field_parameter("center")
    DW = data.pf.domain_right_edge - data.pf.domain_left_edge
    radius = np.zeros(data[field_prefix+"x"].shape, dtype='float64')
    r = radius.copy()
    if any(data.pf.periodicity):
        rdw = radius.copy()
    for i, ax in enumerate('xyz'):
        np.subtract(data["%s%s" % (field_prefix, ax)], center[i], r)
        if data.pf.periodicity[i] == True:
            np.abs(r, r)
            np.subtract(r, DW[i], rdw)
            np.abs(rdw, rdw)
            np.minimum(r, rdw, r)
        np.power(r, 2.0, r)
        np.add(radius, r, radius)
    np.sqrt(radius, radius)
    return radius

def _ParticleRadius(field, data):
    return get_radius(data, "particle_position_")
def _Radius(field, data):
    return get_radius(data, "")

def _ConvertRadiusCGS(data):
    return data.convert("cm")
add_field("ParticleRadius", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusCGS, units=r"\rm{cm}",
          particle_type = True,
          display_name = "Particle Radius")
add_field("Radius", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusCGS, units=r"\rm{cm}")

def _ConvertRadiusMpc(data):
    return data.convert("mpc")
add_field("RadiusMpc", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusMpc, units=r"\rm{Mpc}",
          display_name = "Radius")
add_field("ParticleRadiusMpc", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusMpc, units=r"\rm{Mpc}",
          particle_type=True,
          display_name = "Particle Radius")

def _ConvertRadiuskpc(data):
    return data.convert("kpc")
add_field("ParticleRadiuskpc", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuskpc, units=r"\rm{kpc}",
          particle_type=True,
          display_name = "Particle Radius")
add_field("Radiuskpc", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuskpc, units=r"\rm{kpc}",
          display_name = "Radius")

def _ConvertRadiuskpch(data):
    return data.convert("kpch")
add_field("ParticleRadiuskpch", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuskpch, units=r"\rm{kpc}/\rm{h}",
          particle_type=True,
          display_name = "Particle Radius")
add_field("Radiuskpch", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuskpc, units=r"\rm{kpc}/\rm{h}",
          display_name = "Radius")

def _ConvertRadiuspc(data):
    return data.convert("pc")
add_field("ParticleRadiuspc", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuspc, units=r"\rm{pc}",
          particle_type=True,
          display_name = "Particle Radius")
add_field("Radiuspc", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiuspc, units=r"\rm{pc}",
          display_name="Radius")

def _ConvertRadiusAU(data):
    return data.convert("au")
add_field("ParticleRadiusAU", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusAU, units=r"\rm{AU}",
          particle_type=True,
          display_name = "Particle Radius")
add_field("RadiusAU", function=_Radius,
          validators=[ValidateParameter("center")],
          convert_function = _ConvertRadiusAU, units=r"\rm{AU}",
          display_name = "Radius")

add_field("ParticleRadiusCode", function=_ParticleRadius,
          validators=[ValidateParameter("center")],
          particle_type=True,
          display_name = "Particle Radius (code)")
add_field("RadiusCode", function=_Radius,
          validators=[ValidateParameter("center")],
          display_name = "Radius (code)")

def _RadialVelocity(field, data):
    normal = data.get_field_parameter("normal")
    velocities = obtain_rv_vec(data)    
    theta = data['sph_theta']
    phi   = data['sph_phi']

    return get_sph_r_component(velocities, theta, phi, normal)

def _RadialVelocityABS(field, data):
    return np.abs(_RadialVelocity(field, data))
def _ConvertRadialVelocityKMS(data):
    return km_per_cm
add_field("RadialVelocity", function=_RadialVelocity,
          units=r"\rm{cm}/\rm{s}")
add_field("RadialVelocityABS", function=_RadialVelocityABS,
          units=r"\rm{cm}/\rm{s}")
add_field("RadialVelocityKMS", function=_RadialVelocity,
          convert_function=_ConvertRadialVelocityKMS, units=r"\rm{km}/\rm{s}")
add_field("RadialVelocityKMSABS", function=_RadialVelocityABS,
          convert_function=_ConvertRadialVelocityKMS, units=r"\rm{km}/\rm{s}")

def _TangentialVelocity(field, data):
    return np.sqrt(data["VelocityMagnitude"]**2.0
                 - data["RadialVelocity"]**2.0)
add_field("TangentialVelocity", 
          function=_TangentialVelocity,
          take_log=False, units=r"\rm{cm}/\rm{s}")

def _CuttingPlaneVelocityX(field, data):
    x_vec, y_vec, z_vec = [data.get_field_parameter("cp_%s_vec" % (ax))
                           for ax in 'xyz']
    bulk_velocity = data.get_field_parameter("bulk_velocity")
    if bulk_velocity == None:
        bulk_velocity = np.zeros(3)
    v_vec = np.array([data["%s-velocity" % ax] for ax in 'xyz']) \
                - bulk_velocity[...,np.newaxis]
    return np.dot(x_vec, v_vec)
add_field("CuttingPlaneVelocityX", 
          function=_CuttingPlaneVelocityX,
          validators=[ValidateParameter("cp_%s_vec" % ax)
                      for ax in 'xyz'], units=r"\rm{km}/\rm{s}")
def _CuttingPlaneVelocityY(field, data):
    x_vec, y_vec, z_vec = [data.get_field_parameter("cp_%s_vec" % (ax))
                           for ax in 'xyz']
    bulk_velocity = data.get_field_parameter("bulk_velocity")
    if bulk_velocity == None:
        bulk_velocity = np.zeros(3)
    v_vec = np.array([data["%s-velocity" % ax] for ax in 'xyz']) \
                - bulk_velocity[...,np.newaxis]
    return np.dot(y_vec, v_vec)
add_field("CuttingPlaneVelocityY", 
          function=_CuttingPlaneVelocityY,
          validators=[ValidateParameter("cp_%s_vec" % ax)
                      for ax in 'xyz'], units=r"\rm{km}/\rm{s}")

def _CuttingPlaneBx(field, data):
    x_vec, y_vec, z_vec = [data.get_field_parameter("cp_%s_vec" % (ax))
                           for ax in 'xyz']
    b_vec = np.array([data["B%s" % ax] for ax in 'xyz'])
    return np.dot(x_vec, b_vec)
add_field("CuttingPlaneBx", 
          function=_CuttingPlaneBx,
          validators=[ValidateParameter("cp_%s_vec" % ax)
                      for ax in 'xyz'], units=r"\rm{Gauss}")
def _CuttingPlaneBy(field, data):
    x_vec, y_vec, z_vec = [data.get_field_parameter("cp_%s_vec" % (ax))
                           for ax in 'xyz']
    b_vec = np.array([data["B%s" % ax] for ax in 'xyz'])
    return np.dot(y_vec, b_vec)
add_field("CuttingPlaneBy", 
          function=_CuttingPlaneBy,
          validators=[ValidateParameter("cp_%s_vec" % ax)
                      for ax in 'xyz'], units=r"\rm{Gauss}")

def _MeanMolecularWeight(field,data):
    return (data["Density"] / (mh *data["NumberDensity"]))
add_field("MeanMolecularWeight",function=_MeanMolecularWeight,units=r"")

def _JeansMassMsun(field,data):
    MJ_constant = (((5*kboltz)/(G*mh))**(1.5)) * \
    (3/(4*3.1415926535897931))**(0.5) / 1.989e33

    return (MJ_constant *
            ((data["Temperature"]/data["MeanMolecularWeight"])**(1.5)) *
            (data["Density"]**(-0.5)))
add_field("JeansMassMsun",function=_JeansMassMsun,
          units=r"\rm{M_{\odot}}")

# We add these fields so that the field detector can use them
for field in ["particle_position_%s" % ax for ax in "xyz"] + \
             ["ParticleMass"]:
    # This marker should let everyone know not to use the fields, but NullFunc
    # should do that, too.
    add_field(field, function=NullFunc, particle_type = True,
        units=r"UNDEFINED")

def _pdensity(field, data):
    blank = np.zeros(data.ActiveDimensions, dtype='float64')
    if data["particle_position_x"].size == 0: return blank
    CICDeposit_3(data["particle_position_x"].astype(np.float64),
                 data["particle_position_y"].astype(np.float64),
                 data["particle_position_z"].astype(np.float64),
                 data["ParticleMass"],
                 data["particle_position_x"].size,
                 blank, np.array(data.LeftEdge).astype(np.float64),
                 np.array(data.ActiveDimensions).astype(np.int32),
                 np.float64(data['dx']))
    np.divide(blank, data["CellVolume"], blank)
    return blank
add_field("particle_density", function=_pdensity,
          validators=[ValidateGridType()],
          display_name=r"\mathrm{Particle}\/\mathrm{Density}")

def _MagneticEnergy(field,data):
    """This assumes that your front end has provided Bx, By, Bz in
    units of Gauss. If you use MKS, make sure to write your own
    MagneticEnergy field to deal with non-unitary \mu_0.
    """
    return (data["Bx"]**2 + data["By"]**2 + data["Bz"]**2)/(8*np.pi)
add_field("MagneticEnergy",function=_MagneticEnergy,
          units=r"\rm{ergs}\/\rm{cm}^{-3}",
          display_name=r"\rm{Magnetic}\/\rm{Energy}")

def _BMagnitude(field,data):
    """This assumes that your front end has provided Bx, By, Bz in
    units of Gauss. If you use MKS, make sure to write your own
    BMagnitude field to deal with non-unitary \mu_0.
    """
    return np.sqrt((data["Bx"]**2 + data["By"]**2 + data["Bz"]**2))
add_field("BMagnitude",
          function=_BMagnitude,
          display_name=r"|B|", units=r"\rm{Gauss}")

def _PlasmaBeta(field,data):
    """This assumes that your front end has provided Bx, By, Bz in
    units of Gauss. If you use MKS, make sure to write your own
    PlasmaBeta field to deal with non-unitary \mu_0.
    """
    return data['Pressure']/data['MagneticEnergy']
add_field("PlasmaBeta",
          function=_PlasmaBeta,
          display_name=r"\rm{Plasma}\/\beta", units="")

def _MagneticPressure(field,data):
    return data['MagneticEnergy']
add_field("MagneticPressure",
          function=_MagneticPressure,
          display_name=r"\rm{Magnetic}\/\rm{Pressure}",
          units=r"\rm{ergs}\/\rm{cm}^{-3}")

def _BPoloidal(field,data):
    normal = data.get_field_parameter("normal")

    Bfields = np.array([data['Bx'], data['By'], data['Bz']])

    theta = data['sph_theta']
    phi   = data['sph_phi']

    return get_sph_theta_component(Bfields, theta, phi, normal)

add_field("BPoloidal", function=_BPoloidal,
          units=r"\rm{Gauss}",
          validators=[ValidateParameter("normal")])

def _BToroidal(field,data):
    normal = data.get_field_parameter("normal")

    Bfields = np.array([data['Bx'], data['By'], data['Bz']])

    phi   = data['sph_phi']

    return get_sph_phi_component(Bfields, phi, normal)

add_field("BToroidal", function=_BToroidal,
          units=r"\rm{Gauss}",
          validators=[ValidateParameter("normal")])

def _BRadial(field,data):
    normal = data.get_field_parameter("normal")

    Bfields = np.array([data['Bx'], data['By'], data['Bz']])

    theta = data['sph_theta']
    phi   = data['sph_phi']

    return get_sph_r_component(Bfields, theta, phi, normal)

add_field("BRadial", function=_BRadial,
          units=r"\rm{Gauss}",
          validators=[ValidateParameter("normal")])

def _VorticitySquared(field, data):
    mylog.debug("Generating vorticity on %s", data)
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["x-velocity"].shape)
    dvzdy = (data["z-velocity"][1:-1,sl_right,1:-1] -
             data["z-velocity"][1:-1,sl_left,1:-1]) \
             / (div_fac*data["dy"].flat[0])
    dvydz = (data["y-velocity"][1:-1,1:-1,sl_right] -
             data["y-velocity"][1:-1,1:-1,sl_left]) \
             / (div_fac*data["dz"].flat[0])
    new_field[1:-1,1:-1,1:-1] += (dvzdy - dvydz)**2.0
    del dvzdy, dvydz
    dvxdz = (data["x-velocity"][1:-1,1:-1,sl_right] -
             data["x-velocity"][1:-1,1:-1,sl_left]) \
             / (div_fac*data["dz"].flat[0])
    dvzdx = (data["z-velocity"][sl_right,1:-1,1:-1] -
             data["z-velocity"][sl_left,1:-1,1:-1]) \
             / (div_fac*data["dx"].flat[0])
    new_field[1:-1,1:-1,1:-1] += (dvxdz - dvzdx)**2.0
    del dvxdz, dvzdx
    dvydx = (data["y-velocity"][sl_right,1:-1,1:-1] -
             data["y-velocity"][sl_left,1:-1,1:-1]) \
             / (div_fac*data["dx"].flat[0])
    dvxdy = (data["x-velocity"][1:-1,sl_right,1:-1] -
             data["x-velocity"][1:-1,sl_left,1:-1]) \
             / (div_fac*data["dy"].flat[0])
    new_field[1:-1,1:-1,1:-1] += (dvydx - dvxdy)**2.0
    del dvydx, dvxdy
    new_field = np.abs(new_field)
    return new_field
def _convertVorticitySquared(data):
    return data.convert("cm")**-2.0
add_field("VorticitySquared", function=_VorticitySquared,
          validators=[ValidateSpatial(1,
              ["x-velocity","y-velocity","z-velocity"])],
          units=r"\rm{s}^{-2}",
          convert_function=_convertVorticitySquared)

def _gradPressureX(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Pressure"].shape, dtype='float64')
    ds = div_fac * data['dx'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Pressure"][sl_right,1:-1,1:-1]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Pressure"][sl_left ,1:-1,1:-1]/ds
    return new_field
def _gradPressureY(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Pressure"].shape, dtype='float64')
    ds = div_fac * data['dy'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Pressure"][1:-1,sl_right,1:-1]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Pressure"][1:-1,sl_left ,1:-1]/ds
    return new_field
def _gradPressureZ(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Pressure"].shape, dtype='float64')
    ds = div_fac * data['dz'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Pressure"][1:-1,1:-1,sl_right]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Pressure"][1:-1,1:-1,sl_left ]/ds
    return new_field
def _convertgradPressure(data):
    return 1.0/data.convert("cm")
for ax in 'XYZ':
    n = "gradPressure%s" % ax
    add_field(n, function=eval("_%s" % n),
              convert_function=_convertgradPressure,
              validators=[ValidateSpatial(1, ["Pressure"])],
              units=r"\rm{dyne}/\rm{cm}^{3}")

def _gradPressureMagnitude(field, data):
    return np.sqrt(data["gradPressureX"]**2 +
                   data["gradPressureY"]**2 +
                   data["gradPressureZ"]**2)
add_field("gradPressureMagnitude", function=_gradPressureMagnitude,
          validators=[ValidateSpatial(1, ["Pressure"])],
          units=r"\rm{dyne}/\rm{cm}^{3}")

def _gradDensityX(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Density"].shape, dtype='float64')
    ds = div_fac * data['dx'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Density"][sl_right,1:-1,1:-1]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Density"][sl_left ,1:-1,1:-1]/ds
    return new_field
def _gradDensityY(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Density"].shape, dtype='float64')
    ds = div_fac * data['dy'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Density"][1:-1,sl_right,1:-1]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Density"][1:-1,sl_left ,1:-1]/ds
    return new_field
def _gradDensityZ(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["Density"].shape, dtype='float64')
    ds = div_fac * data['dz'].flat[0]
    new_field[1:-1,1:-1,1:-1]  = data["Density"][1:-1,1:-1,sl_right]/ds
    new_field[1:-1,1:-1,1:-1] -= data["Density"][1:-1,1:-1,sl_left ]/ds
    return new_field
def _convertgradDensity(data):
    return 1.0/data.convert("cm")
for ax in 'XYZ':
    n = "gradDensity%s" % ax
    add_field(n, function=eval("_%s" % n),
              convert_function=_convertgradDensity,
              validators=[ValidateSpatial(1, ["Density"])],
              units=r"\rm{g}/\rm{cm}^{4}")

def _gradDensityMagnitude(field, data):
    return np.sqrt(data["gradDensityX"]**2 +
                   data["gradDensityY"]**2 +
                   data["gradDensityZ"]**2)
add_field("gradDensityMagnitude", function=_gradDensityMagnitude,
          validators=[ValidateSpatial(1, ["Density"])],
          units=r"\rm{g}/\rm{cm}^{4}")

def _BaroclinicVorticityX(field, data):
    rho2 = data["Density"].astype('float64')**2
    return (data["gradPressureY"] * data["gradDensityZ"] -
            data["gradPressureZ"] * data["gradDensityY"]) / rho2
def _BaroclinicVorticityY(field, data):
    rho2 = data["Density"].astype('float64')**2
    return (data["gradPressureZ"] * data["gradDensityX"] -
            data["gradPressureX"] * data["gradDensityZ"]) / rho2
def _BaroclinicVorticityZ(field, data):
    rho2 = data["Density"].astype('float64')**2
    return (data["gradPressureX"] * data["gradDensityY"] -
            data["gradPressureY"] * data["gradDensityX"]) / rho2
for ax in 'XYZ':
    n = "BaroclinicVorticity%s" % ax
    add_field(n, function=eval("_%s" % n),
          validators=[ValidateSpatial(1, ["Density", "Pressure"])],
          units=r"\rm{s}^{-1}")

def _BaroclinicVorticityMagnitude(field, data):
    return np.sqrt(data["BaroclinicVorticityX"]**2 +
                   data["BaroclinicVorticityY"]**2 +
                   data["BaroclinicVorticityZ"]**2)
add_field("BaroclinicVorticityMagnitude",
          function=_BaroclinicVorticityMagnitude,
          validators=[ValidateSpatial(1, ["Density", "Pressure"])],
          units=r"\rm{s}^{-1}")

def _VorticityX(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["z-velocity"].shape, dtype='float64')
    new_field[1:-1,1:-1,1:-1] = (data["z-velocity"][1:-1,sl_right,1:-1] -
                                 data["z-velocity"][1:-1,sl_left,1:-1]) \
                                 / (div_fac*data["dy"].flat[0])
    new_field[1:-1,1:-1,1:-1] -= (data["y-velocity"][1:-1,1:-1,sl_right] -
                                  data["y-velocity"][1:-1,1:-1,sl_left]) \
                                  / (div_fac*data["dz"].flat[0])
    return new_field
def _VorticityY(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["z-velocity"].shape, dtype='float64')
    new_field[1:-1,1:-1,1:-1] = (data["x-velocity"][1:-1,1:-1,sl_right] -
                                 data["x-velocity"][1:-1,1:-1,sl_left]) \
                                 / (div_fac*data["dz"].flat[0])
    new_field[1:-1,1:-1,1:-1] -= (data["z-velocity"][sl_right,1:-1,1:-1] -
                                  data["z-velocity"][sl_left,1:-1,1:-1]) \
                                  / (div_fac*data["dx"].flat[0])
    return new_field
def _VorticityZ(field, data):
    # We need to set up stencils
    if data.pf["HydroMethod"] == 2:
        sl_left = slice(None,-2,None)
        sl_right = slice(1,-1,None)
        div_fac = 1.0
    else:
        sl_left = slice(None,-2,None)
        sl_right = slice(2,None,None)
        div_fac = 2.0
    new_field = np.zeros(data["x-velocity"].shape, dtype='float64')
    new_field[1:-1,1:-1,1:-1] = (data["y-velocity"][sl_right,1:-1,1:-1] -
                                 data["y-velocity"][sl_left,1:-1,1:-1]) \
                                 / (div_fac*data["dx"].flat[0])
    new_field[1:-1,1:-1,1:-1] -= (data["x-velocity"][1:-1,sl_right,1:-1] -
                                  data["x-velocity"][1:-1,sl_left,1:-1]) \
                                  / (div_fac*data["dy"].flat[0])
    return new_field
def _convertVorticity(data):
    return 1.0/data.convert("cm")
for ax in 'XYZ':
    n = "Vorticity%s" % ax
    add_field(n, function=eval("_%s" % n),
              convert_function=_convertVorticity,
              validators=[ValidateSpatial(1, 
                          ["x-velocity", "y-velocity", "z-velocity"])],
              units=r"\rm{s}^{-1}")

def _VorticityMagnitude(field, data):
    return np.sqrt(data["VorticityX"]**2 +
                   data["VorticityY"]**2 +
                   data["VorticityZ"]**2)
add_field("VorticityMagnitude", function=_VorticityMagnitude,
          validators=[ValidateSpatial(1, 
                      ["x-velocity", "y-velocity", "z-velocity"])],
          units=r"\rm{s}^{-1}")

def _VorticityStretchingX(field, data):
    return data["DivV"] * data["VorticityX"]
def _VorticityStretchingY(field, data):
    return data["DivV"] * data["VorticityY"]
def _VorticityStretchingZ(field, data):
    return data["DivV"] * data["VorticityZ"]
for ax in 'XYZ':
    n = "VorticityStretching%s" % ax
    add_field(n, function=eval("_%s" % n),
              validators=[ValidateSpatial(0)])
def _VorticityStretchingMagnitude(field, data):
    return np.sqrt(data["VorticityStretchingX"]**2 +
                   data["VorticityStretchingY"]**2 +
                   data["VorticityStretchingZ"]**2)
add_field("VorticityStretchingMagnitude", 
          function=_VorticityStretchingMagnitude,
          validators=[ValidateSpatial(1, 
                      ["x-velocity", "y-velocity", "z-velocity"])],
          units=r"\rm{s}^{-1}")

def _VorticityGrowthX(field, data):
    return -data["VorticityStretchingX"] - data["BaroclinicVorticityX"]
def _VorticityGrowthY(field, data):
    return -data["VorticityStretchingY"] - data["BaroclinicVorticityY"]
def _VorticityGrowthZ(field, data):
    return -data["VorticityStretchingZ"] - data["BaroclinicVorticityZ"]
for ax in 'XYZ':
    n = "VorticityGrowth%s" % ax
    add_field(n, function=eval("_%s" % n),
              validators=[ValidateSpatial(1, 
                          ["x-velocity", "y-velocity", "z-velocity"])],
              units=r"\rm{s}^{-2}")
def _VorticityGrowthMagnitude(field, data):
    result = np.sqrt(data["VorticityGrowthX"]**2 +
                     data["VorticityGrowthY"]**2 +
                     data["VorticityGrowthZ"]**2)
    dot = np.zeros(result.shape)
    for ax in "XYZ":
        dot += data["Vorticity%s" % ax] * data["VorticityGrowth%s" % ax]
    result = np.sign(dot) * result
    return result
add_field("VorticityGrowthMagnitude", function=_VorticityGrowthMagnitude,
          validators=[ValidateSpatial(1, 
                      ["x-velocity", "y-velocity", "z-velocity"])],
          units=r"\rm{s}^{-1}",
          take_log=False)
def _VorticityGrowthMagnitudeABS(field, data):
    return np.sqrt(data["VorticityGrowthX"]**2 +
                   data["VorticityGrowthY"]**2 +
                   data["VorticityGrowthZ"]**2)
add_field("VorticityGrowthMagnitudeABS", function=_VorticityGrowthMagnitudeABS,
          validators=[ValidateSpatial(1, 
                      ["x-velocity", "y-velocity", "z-velocity"])],
          units=r"\rm{s}^{-1}")

def _VorticityGrowthTimescale(field, data):
    domegax_dt = data["VorticityX"] / data["VorticityGrowthX"]
    domegay_dt = data["VorticityY"] / data["VorticityGrowthY"]
    domegaz_dt = data["VorticityZ"] / data["VorticityGrowthZ"]
    return np.sqrt(domegax_dt**2 + domegay_dt**2 + domegaz_dt)
add_field("VorticityGrowthTimescale", function=_VorticityGrowthTimescale,
          validators=[ValidateSpatial(1, 
                      ["x-velocity", "y-velocity", "z-velocity"])],
          units=r"\rm{s}")

########################################################################
# With radiation pressure
########################################################################

def _VorticityRadPressureX(field, data):
    rho = data["Density"].astype('float64')
    return (data["RadAccel2"] * data["gradDensityZ"] -
            data["RadAccel3"] * data["gradDensityY"]) / rho
def _VorticityRadPressureY(field, data):
    rho = data["Density"].astype('float64')
    return (data["RadAccel3"] * data["gradDensityX"] -
            data["RadAccel1"] * data["gradDensityZ"]) / rho
def _VorticityRadPressureZ(field, data):
    rho = data["Density"].astype('float64')
    return (data["RadAccel1"] * data["gradDensityY"] -
            data["RadAccel2"] * data["gradDensityX"]) / rho
def _convertRadAccel(data):
    return data.convert("x-velocity")/data.convert("Time")
for ax in 'XYZ':
    n = "VorticityRadPressure%s" % ax
    add_field(n, function=eval("_%s" % n),
              convert_function=_convertRadAccel,
              validators=[ValidateSpatial(1, 
                   ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
              units=r"\rm{s}^{-1}")

def _VorticityRadPressureMagnitude(field, data):
    return np.sqrt(data["VorticityRadPressureX"]**2 +
                   data["VorticityRadPressureY"]**2 +
                   data["VorticityRadPressureZ"]**2)
add_field("VorticityRadPressureMagnitude",
          function=_VorticityRadPressureMagnitude,
          validators=[ValidateSpatial(1, 
                      ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
          units=r"\rm{s}^{-1}")

def _VorticityRPGrowthX(field, data):
    return -data["VorticityStretchingX"] - data["BaroclinicVorticityX"] \
           -data["VorticityRadPressureX"]
def _VorticityRPGrowthY(field, data):
    return -data["VorticityStretchingY"] - data["BaroclinicVorticityY"] \
           -data["VorticityRadPressureY"]
def _VorticityRPGrowthZ(field, data):
    return -data["VorticityStretchingZ"] - data["BaroclinicVorticityZ"] \
           -data["VorticityRadPressureZ"]
for ax in 'XYZ':
    n = "VorticityRPGrowth%s" % ax
    add_field(n, function=eval("_%s" % n),
              validators=[ValidateSpatial(1, 
                       ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
              units=r"\rm{s}^{-1}")
def _VorticityRPGrowthMagnitude(field, data):
    result = np.sqrt(data["VorticityRPGrowthX"]**2 +
                     data["VorticityRPGrowthY"]**2 +
                     data["VorticityRPGrowthZ"]**2)
    dot = np.zeros(result.shape)
    for ax in "XYZ":
        dot += data["Vorticity%s" % ax] * data["VorticityGrowth%s" % ax]
    result = np.sign(dot) * result
    return result
add_field("VorticityRPGrowthMagnitude", function=_VorticityGrowthMagnitude,
          validators=[ValidateSpatial(1, 
                      ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
          units=r"\rm{s}^{-1}",
          take_log=False)
def _VorticityRPGrowthMagnitudeABS(field, data):
    return np.sqrt(data["VorticityRPGrowthX"]**2 +
                   data["VorticityRPGrowthY"]**2 +
                   data["VorticityRPGrowthZ"]**2)
add_field("VorticityRPGrowthMagnitudeABS", 
          function=_VorticityRPGrowthMagnitudeABS,
          validators=[ValidateSpatial(1, 
                      ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
          units=r"\rm{s}^{-1}")

def _VorticityRPGrowthTimescale(field, data):
    domegax_dt = data["VorticityX"] / data["VorticityRPGrowthX"]
    domegay_dt = data["VorticityY"] / data["VorticityRPGrowthY"]
    domegaz_dt = data["VorticityZ"] / data["VorticityRPGrowthZ"]
    return np.sqrt(domegax_dt**2 + domegay_dt**2 + domegaz_dt**2)
add_field("VorticityRPGrowthTimescale", function=_VorticityRPGrowthTimescale,
          validators=[ValidateSpatial(1, 
                      ["Density", "RadAccel1", "RadAccel2", "RadAccel3"])],
          units=r"\rm{s}^{-1}")



"""
Easy plotting.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from _mpl_imports import *
from yt.data_objects.profiles import BinnedProfile1D

plot_type_registry = {}

class EasyPlot(object):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if hasattr(cls, "_type_name"):
                plot_type_registry[cls._type_name] = cls

    def __init__(self):
        pass

    def save(self, fn):
        canvas = FigureCanvasAgg(self.figure)
        canvas.print_figure(fn)

class EasyPDFPlot(EasyPlot):
    _type_name = "PDF"

    def __init__(self, data_source, x_field, x_log = True,
                 n_bins = 128, y_log = True,
                 plot_args = None,
                 figure_args = None):
        if plot_args is None: plot_args = {}
        if figure_args is None: figure_args = {}
        self.x_field = x_field
        self.data_source = data_source
        # Now we just make the plot
        x_min, x_max = self.data_source.quantities["Extrema"](
                x_field, non_zero = x_log, lazy_reader = True)[0]
        self.profile = BinnedProfile1D(self.data_source,
            n_bins, self.x_field, x_min, x_max, x_log,
            lazy_reader = True)
        self.profile.add_fields(["CellMassMsun"], weight=None)
        self.profile["CellMassMsun"] /= self.profile["CellMassMsun"].sum()
        self.figure = matplotlib.figure.Figure(**figure_args)
        self.axes = self.figure.add_subplot(1,1,1)
        if y_log and x_log: f = self.axes.loglog
        elif y_log: f = self.axes.semilogy
        elif x_log: f = self.axes.semilogx
        else: f = self.axes.plot
        self.plot = f(self.profile[x_field], self.profile["CellMassMsun"],
                      **plot_args)
        self.axes.set_xlabel(data_source.pf.field_info[x_field].get_label())



"""
Wrapping code for voro++

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from cython.operator cimport dereference as deref, preincrement as inc
from libc.stdlib cimport malloc, free, abs, calloc, labs
cimport libcpp

import numpy as np
cimport numpy as np
cimport cython

cdef extern from "voro++.cc":
    cdef cppclass container:
        container(double xmin, double xmax, double ymin, double ymax,
                  double zmin, double zmax, int nx, int ny, int nz,
                  libcpp.bool xper, libcpp.bool yper, libcpp.bool zper, int alloc)
        void put(int n, double x, double y, double z)
        void store_cell_volumes(double *vols)

cdef class VoronoiVolume:
    cdef container *my_con
    cdef int npart
    def __init__(self, xi, yi, zi):
        self.my_con = new container(0.0, 1.0, 0.0, 1.0, 0.0, 1.0,
                                    xi, yi, zi, False, False, False, 8)
        self.npart = 0

    def __dealloc__(self):
        del self.my_con

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_array(self, np.ndarray[np.float64_t, ndim=1] xpos,
                        np.ndarray[np.float64_t, ndim=1] ypos,
                        np.ndarray[np.float64_t, ndim=1] zpos):
        cdef int i
        for i in range(xpos.shape[0]):
            self.my_con.put(self.npart, xpos[i], ypos[i], zpos[i])
            self.npart += 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_volumes(self):
        cdef np.ndarray vol = np.zeros(self.npart, 'double')
        cdef double *vdouble = <double *> vol.data
        self.my_con.store_cell_volumes(vdouble)
        return vol



#
# Physical Constants and Units Conversion Factors
#
# Values for these constants are drawn from IAU and IUPAC data 
# unless otherwise noted:
# http://maia.usno.navy.mil/NSFA/IAU2009_consts.html
# http://goldbook.iupac.org/list_goldbook_phys_constants_defs.html

# Masses
mass_hydrogen_cgs = 1.674534e-24  # g
mass_electron_cgs = 9.1093898e-28  # g
amu_cgs           = 1.6605402e-24  # g
mass_sun_cgs = 1.98841586e33  # g
# Velocities
speed_of_light_cgs = 2.99792458e10  # cm/s, exact

# Cross Sections
# 8*pi/3 (alpha*hbar*c/(2*pi))**2
cross_section_thompson_cgs = 6.65245854533e-25  # cm^2

# Charge
charge_proton_cgs = 4.8032056e-10  # esu = 1.602176487e-19  Coulombs

# Physical Constants
boltzmann_constant_cgs = 1.3806504e-16  # erg K^-1
gravitational_constant_cgs  = 6.67428e-8  # cm^3 g^-1 s^-2
planck_constant_cgs   = 6.62606896e-27  # erg s
stefan_boltzmann_constant_cgs = 5.67051e-5 # erg cm^-2 s^-1 K^-4
rho_crit_now = 1.8788e-29  # g times h^2 (critical mass for closure, Cosmology)

# Misc. Approximations
mass_mean_atomic_cosmology = 1.22
mass_mean_atomic_galactic = 2.3

# Conversion Factors:  X au * mpc_per_au = Y mpc
# length
mpc_per_mpc   = 1e0
mpc_per_kpc   = 1e-3
mpc_per_pc    = 1e-6
mpc_per_au    = 4.84813682e-12
mpc_per_rsun  = 2.253962e-14
mpc_per_miles = 5.21552871e-20
mpc_per_km    = 3.24077929e-20
mpc_per_cm    = 3.24077929e-25
kpc_per_cm    = mpc_per_cm / mpc_per_kpc
km_per_pc     = 3.08567758e13
km_per_m      = 1e-3
km_per_cm     = 1e-5
pc_per_cm     = 3.24077929e-19

m_per_fpc     = 0.0324077929

kpc_per_mpc   = 1.0 / mpc_per_kpc
pc_per_mpc    = 1.0 / mpc_per_pc
au_per_mpc    = 1.0 / mpc_per_au
rsun_per_mpc  = 1.0 / mpc_per_rsun
miles_per_mpc = 1.0 / mpc_per_miles
km_per_mpc    = 1.0 / mpc_per_km
cm_per_mpc    = 1.0 / mpc_per_cm
cm_per_kpc    = 1.0 / kpc_per_cm
cm_per_km     = 1.0 / km_per_cm
pc_per_km     = 1.0 / km_per_pc
cm_per_pc     = 1.0 / pc_per_cm

# time
sec_per_Gyr  = 31.5576e15
sec_per_Myr  = 31.5576e12
sec_per_kyr  = 31.5576e9
sec_per_year = 31.5576e6   # "IAU Style Manual" by G.A. Wilkins, Comm. 5, in IAU Transactions XXB (1989)
sec_per_day  = 86400.0
sec_per_hr   = 3600.0
day_per_year = 365.25

# temperature / energy
erg_per_eV = 1.602176487e-12 # http://goldbook.iupac.org/E02014.html
erg_per_keV = erg_per_eV * 1.0e3
K_per_keV = erg_per_keV / boltzmann_constant_cgs
keV_per_K = 1.0 / K_per_keV

#Short cuts
G = gravitational_constant_cgs
me = mass_electron_cgs
mp = mass_hydrogen_cgs
qp = charge_proton_cgs
mh = mp
clight = speed_of_light_cgs
kboltz = boltzmann_constant_cgs
hcgs = planck_constant_cgs
sigma_thompson = cross_section_thompson_cgs
Na = 1 / amu_cgs



"""
Transformations between coordinate systems

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: JS Oishi <jsoishi@astro.berkeley.edu>
Organization: UC Berkeley
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk, J. S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from yt.funcs import *

from yt.utilities.linear_interpolators import \
    TrilinearFieldInterpolator

def spherical_regrid(pf, nr, ntheta, nphi, rmax, fields,
                     center=None, smoothed=True):
    """
    This function takes a parameter file (*pf*) along with the *nr*, *ntheta*
    and *nphi* points to generate out to *rmax*, and it grids *fields* onto
    those points and returns a dict.  *center* if supplied will be the center,
    otherwise the most dense point will be chosen.  *smoothed* governs whether
    regular covering grids or smoothed covering grids will be used.
    """
    mylog.warning("This code may produce some artifacts of interpolation")
    mylog.warning("See yt/extensions/coordinate_transforms.py for plotting information")
    if center is None: center = pf.h.find_max("Density")[1]
    fields = ensure_list(fields)
    r,theta,phi = np.mgrid[0:rmax:nr*1j,
                           0:np.pi:ntheta*1j,
                           0:2*np.pi:nphi*1j]
    new_grid = dict(r=r, theta=theta, phi=phi)
    new_grid['x'] = r*np.sin(theta)*np.cos(phi) + center[0]
    new_grid['y'] = r*np.sin(theta)*np.sin(phi) + center[1]
    new_grid['z'] = r*np.cos(theta)             + center[2]
    sphere = pf.h.sphere(center, rmax)
    return arbitrary_regrid(new_grid, sphere, fields, smoothed)

def arbitrary_regrid(new_grid, data_source, fields, smoothed=True):
    """
    This function accepts a dict of points 'x', 'y' and 'z' and a data source
    from which to interpolate new points, along with a list of fields it needs
    to regrid onto those xyz points.  It then returns interpolated points.
    This has not been well-tested other than for regular spherical regridding.
    """
    fields = ensure_list(fields)
    new_grid['handled'] = np.zeros(new_grid['x'].shape, dtype='bool')
    for field in fields:
        new_grid[field] = np.zeros(new_grid['x'].shape, dtype='float64')
    grid_order = np.argsort(data_source.gridLevels)
    ng = len(data_source._grids)

    for i,grid in enumerate(data_source._grids[grid_order][::-1]):
        mylog.info("Regridding grid % 4i / % 4i (%s - %s)", i, ng, grid.id, grid.Level)
        cg = grid.retrieve_ghost_zones(1, fields, smoothed=smoothed)

        # makes x0,x1,y0,y1,z0,z1
        bounds = np.concatenate(zip(cg.left_edge, cg.right_edge)) 

        
        # Now we figure out which of our points are inside this grid
        # Note that we're only looking at the grid, not the grid-with-ghost-zones
        point_ind = np.ones(new_grid['handled'].shape, dtype='bool') # everything at first
        for i,ax in enumerate('xyz'): # i = 0,1,2 ; ax = x, y, z
            # &= does a logical_and on the array
            point_ind &= ( ( grid.LeftEdge[i] <= new_grid[ax]      )
                         & ( new_grid[ax]     <= grid.RightEdge[i] ) )
        point_ind &= (new_grid['handled'] == False) # only want unhandled points

        # If we don't have any, we can just leave
        if point_ind.sum() == 0: continue

        # because of the funky way the interpolator takes points, we have to make a
        # new dict of just the points inside this grid
        point_grid = {'x' : new_grid['x'][point_ind],
                      'y' : new_grid['y'][point_ind],
                      'z' : new_grid['z'][point_ind]}

        # Now we know which of the points in new_grid are inside this grid
        for field in fields:
            interpolator = TrilinearFieldInterpolator(
                cg[field],bounds,['x','y','z'])
            new_grid[field][point_ind] = interpolator(point_grid)

        new_grid['handled'][point_ind] = True

    mylog.info("Finished with %s dangling points",
        new_grid['handled'].size - new_grid['handled'].sum())

    return new_grid

"""
# The following will work to plot through different slices:

import pylab
for i in range(n_theta):
    print "Doing % 3i / % 3i" % (i, n_theta)
    pylab.clf()
    ax=pylab.subplot(1,1,1, projection="polar", aspect=1.)
    ax.pcolormesh(phi[:,i,:], r[:,i,:],
                  np.log10(sph_grid[field][:,i,:]))
    pylab.savefig("polar/latitude_%03i.png" % i)

for i in range(n_phi):
    print "Doing % 3i / % 3i" % (i, n_phi)
    pylab.clf()
    ax=pylab.subplot(1,1,1, projection="polar", aspect=1.)
    ax.pcolormesh(theta[:,:,i], r[:,:,i],
                  np.log10(sph_grid[field][:,:,i]))
    pylab.savefig("polar/longitude_%03i.png" % i)
"""
